import os
import shutil
from typing import Tuple

from alabamaEncode.conent_analysis.chunk.final_encode_step import (
    FinalEncodeStep,
)
from alabamaEncode.conent_analysis.opinionated_vmaf import get_crf_limits
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.calculate import get_metric_from_stats
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.chunk import ChunkObject


def get_weighed_vmaf_score(
    _stats: EncodeStats,
    codec: Codec,
    statistical_representation: str,
    metric_target: float,
) -> float:
    """
    Score to minimize

    formula so that after ~1Mb/s & 12 frames we start to exponentially add weight,
    taking frames and bitrate into account will make sure that
    short high bitrate blips will mostly get ignored,
    but longer larger bitrates will be penalised
    at about 10Mb/s & 48 frames the weight will be ~5,
    so for example, 10Mb/s for 48 frames will have the same impact as a vmaf error of 5
    The overall goal is get greedy with bitrate while being guided by vmaf,
    because pure vmaf guidance can lead to keeping a lot of unnecessary grain on motionless scenes,
    exacerbated on low quality sources
    """

    if codec != Codec.av1:
        print(f"weighted vmaf is tuned for AV1 codecs, expect suboptimal results")

    metric = get_metric_from_stats(
        _stats, statistical_representation=statistical_representation
    )
    frames = _stats.length_frames
    bitrate = _stats.bitrate  # in kbps

    # bitrate component
    bitrate_weight = (bitrate / 650) ** 0.4 - 1
    bitrate_weight *= 2

    # Exponential component for the number of frames
    size_kb = (frames * bitrate) / 1000
    overall_size_weight = (size_kb / 250) ** 0.5
    overall_size_weight = max(overall_size_weight, 1)

    # Combine the components
    combined_weight = bitrate_weight * overall_size_weight

    # Ensure the combined weight is not negative
    combined_weight = max(combined_weight, 0)
    return abs(metric_target - metric) + combined_weight


class DynamicTargetVmaf(FinalEncodeStep):
    """
    Target vmaf but:
    we add a bitrate weight that was determined using some empirical observation
    now imagine a score, that is the vmaf error + the bitrate weight
    if we plot it against crf we get a U curve or more likely a nike swoosh
    we want to find the minimum of this curve
    binary search does not work because the curve is unimodal and were looking for the minimum.
    Alternatives are ternary search or golden section search
    This is a final encode step, since we will be running all encodes at the slow speed
    and then when we think its good enough we quit.
    We don't run probes at faster speeds because
    I've noticed that the vmaf calculation gives more overhead than actual encoding using reasonable SVTAV1 presets
    """

    def run(
        self, enc: Encoder, chunk: ChunkObject, ctx: AlabamaContext, encoded_a_frame
    ) -> EncodeStats:
        metric_target = ctx.vmaf

        def log(_str):
            ctx.log(
                f"{chunk.log_prefix()}{_str}",
                category="probe",
            )

        original_speed = enc.speed
        original_output_path = enc.output_path
        probe_file_base = ctx.get_probe_file_base(chunk.chunk_path)
        trys = []
        stats = None

        def finish(_stats, crf):
            score_err = get_weighed_vmaf_score(
                stats,
                codec=enc.get_codec(),
                statistical_representation=ctx.vmaf_target_representation,
                metric_target=metric_target,
            )
            log(
                f"Finished with crf {crf}; {metric_name}: "
                f"{get_metric_from_stats(_stats, ctx.vmaf_target_representation)};"
                f" score_error: {score_err}; bitrate: {_stats.bitrate} kb/s"
            )
            ctx.get_kv().set("best_crfs", chunk.chunk_index, crf)
            os.rename(enc.output_path, original_output_path)
            if os.path.exists(probe_file_base):
                shutil.rmtree(probe_file_base)
            return _stats

        metric_name = "vmaf"

        tries = 0
        max_tries = 10

        recent_scores = []

        def run_probe(crf) -> Tuple[float, EncodeStats, float, float, bool]:
            """
            :param crf: crf to try
            :param run_faster_speed: to run at faster speed for initial search probes
            :return: [vmaf error, stats, vmaf]
            """
            enc.crf = crf
            enc.output_path = os.path.join(
                probe_file_base,
                f"{chunk.chunk_index}_{enc.crf}{enc.get_chunk_file_extension()}",
            )
            nonlocal stats
            stats = enc.run(
                metric_to_calculate=Metric.VMAF,
                metric_params=ctx.get_vmaf_options(),
            )
            _metric = get_metric_from_stats(
                stats, statistical_representation=ctx.vmaf_target_representation
            )
            _score = get_weighed_vmaf_score(
                stats,
                codec=enc.get_codec(),
                statistical_representation=ctx.vmaf_target_representation,
                metric_target=metric_target,
            )

            nonlocal tries
            log(
                f"crf: {crf}; {metric_name}: {_metric}; score_error: {_score}; bitrate: {stats.bitrate} kb/s; "
                f"attempt {tries}/{max_tries}"
            )
            trys.append((abs(metric_target - _metric), stats, _metric, _score, crf))

            recent_scores.append(_score)

            # If the list of recent scores is too long, remove the oldest score
            if len(recent_scores) > 4:
                recent_scores.pop(0)

            # Calculate the threshold for the early quit
            threshold = sum(recent_scores) / len(
                recent_scores
            )  # replace this with the desired calculation (e.g., median)

            threshold *= 1.1

            # If the new score is 10% better than the last three ones, quit early
            quit_early = _score < threshold and len(recent_scores) >= 4

            tries += 1
            return abs(metric_target - _metric), stats, _metric, _score, quit_early

        enc.output_path = os.path.join(
            probe_file_base,
            f"{chunk.chunk_index}{enc.get_chunk_file_extension()}",
        )
        enc.passes = 1
        enc.rate_distribution = EncoderRateDistribution.CQ

        best_crf_from_kv = ctx.get_kv().get("best_crfs", chunk.chunk_index)
        if best_crf_from_kv is not None:
            _, stats, _, _, _ = run_probe(best_crf_from_kv)
            return finish(stats, best_crf_from_kv)

        low_crf, high_crf = get_crf_limits(enc, ctx)
        max_score_error = 0.7

        tol = 0.10 if enc.supports_float_crfs() else 1
        gr = (1 + 5**0.5) / 2

        a = high_crf - (high_crf - low_crf) / gr
        b = low_crf + (high_crf - low_crf) / gr

        a = round(a) if not enc.supports_float_crfs() else a
        b = round(b) if not enc.supports_float_crfs() else b

        current_metric_error_a, stats_a, metric_a, score_a, quit_early = run_probe(a)

        if score_a < max_score_error or quit_early:
            return finish(stats_a, a)

        current_metric_error_b, stats_b, metric_b, score_b, quit_early = run_probe(b)

        if score_b < max_score_error or quit_early:
            return finish(stats_b, b)

        while abs(high_crf - low_crf) > tol and tries < max_tries:
            if score_a < score_b:
                high_crf = b
                b = a
                stats_b = stats_a
                score_b = score_a
                current_metric_error_b = current_metric_error_a
                metric_b = metric_a
                a = high_crf - (high_crf - low_crf) / gr
                a = round(a) if not enc.supports_float_crfs() else a

                # check if we already tried this crf
                if any([abs(_x[4] - a) < 0.1 for _x in trys]):
                    log(f"crf {a} already tried, quiting")
                    return finish(stats_a, a)

                (
                    current_metric_error_a,
                    stats_a,
                    metric_a,
                    score_a,
                    quit_early,
                ) = run_probe(a)
                if score_a < max_score_error or quit_early:
                    return finish(stats_a, a)
            else:
                low_crf = a
                a = b
                stats_a = stats_b
                metric_a = metric_b
                score_a = score_b
                b = low_crf + (high_crf - low_crf) / gr
                b = round(b) if not enc.supports_float_crfs() else b

                # check if we already tried this crf
                if any([abs(_x[4] - b) < 0.1 for _x in trys]):
                    log(f"crf {b} already tried, quiting")
                    return finish(stats_b, b)

                (
                    current_metric_error_b,
                    stats_b,
                    metric_b,
                    score_b,
                    quit_early,
                ) = run_probe(b)
                if score_b < max_score_error or quit_early:
                    return finish(stats_b, b)

            # Check if the probe score is within the error
            if max_score_error > score_a or max_score_error > score_b:
                # Log the result and quit early
                log(
                    f"Probe score({min(score_a, score_b)}) within error ({metric_name} "
                    f"err: {min(current_metric_error_a, current_metric_error_b)})"
                    f" {min(stats_a.bitrate, stats_b.bitrate)} kb/s, quiting early"
                )
                # Return the result
                if score_a < score_b:
                    return finish(stats_a, a)
                else:
                    return finish(stats_b, b)

        # Return the middle point of the search interval as the answer
        last_crf_try = (low_crf + high_crf) / 2
        last_crf_try = (
            round(last_crf_try) if not enc.supports_float_crfs() else last_crf_try
        )

        # check if we already tried this crf
        if any([abs(_x[4] - last_crf_try) < 0.1 for _x in trys]):
            log(f"crf {last_crf_try} already tried, quiting")
            return finish(stats_a, last_crf_try)

        current_metric_error, stats, metric, score = run_probe(last_crf_try)
        log(
            f"crf: {last_crf_try} {metric_name}: {metric} score_error: {score} "
            f"attempt {tries}/{max_tries}"
        )
        return finish(stats, last_crf_try)

    def dry_run(self, enc: Encoder, chunk: ChunkObject) -> str:
        raise Exception(f"dry_run not implemented for {self.__class__.__name__}")



from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.scene.chunk import ChunkObject


class PlainVbr(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        enc.rate_distribution = EncoderRateDistribution.VBR
        enc.bitrate = ctx.prototype_encoder.bitrate
        enc.passes = 3
        return enc



import os
import re
import subprocess
import time
from queue import Queue
from threading import Thread
from typing import List, Callable, Optional

__all__ = ["run_cli", "run_cli_parallel", "CliResult"]


class CliResult:
    def __init__(self, return_code, output, time_taken=-1.0):
        self.return_code = return_code
        self.output = output
        self.time_taken = time_taken

    def __repr__(self):
        return f"ExecuteResult(return_code={self.return_code}, output={self.output})"

    def __str__(self):
        return self.output()

    def success(self) -> bool:
        return self.return_code == 0

    def get_output(self) -> str:
        return self.output

    def strip_mp4_warning(self):
        """
        sometimes when using ffprobe on mp4 files,
        [movmp4m4a3gp3g2mj2 @ 0x5579bcaa2fc0] Referenced QT chapter track not found\n
        appears in the output, this function removes it
        :return: self
        """
        self.output = re.sub(r"\[mov.+not found", "", self.output).strip()
        return self

    def verify(
        self,
        fail_message: str = "Cli failed",
        bad_output_hints: List[str] = None,
        files: List[str] = None,
    ):
        """
        :param fail_message: message to raise if the cli failed
        :param bad_output_hints: list of strings that,
        when found in the output, will cause the verification to fail
        :param files: a list of files that must exist and bigger than zero bytes
        :return: self for pipelining
        """
        if not self.success():
            raise RuntimeError(fail_message)
        if bad_output_hints:
            for hint in bad_output_hints:
                if hint in self.output and hint != "":
                    raise RuntimeError(fail_message)
                if hint == self.output:
                    raise RuntimeError(fail_message)
        if files:
            for file in files:
                if not os.path.exists(file) or os.path.getsize(file) == 0:
                    raise RuntimeError(fail_message)
        return self

    def get_as_int(self) -> int:
        return int(self.output.strip())

    def filter_output(self, filter_str: str):
        self.output = self.output.strip().replace(filter_str, "")
        return self

    def get_as_float(self) -> float:
        return float(self.output.strip())


def run_cli(
    cmd,
    timeout_value=-1,
    on_output: Optional[Callable[[str], None]] = None,
) -> CliResult:
    start = time.perf_counter()
    p = subprocess.Popen(
        cmd,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    output = ""
    while p.poll() is None:  # While the process is still running...
        chunk = p.stdout.read(1).decode(errors="ignore")  # Read output chunk by chunk
        output += chunk  # Add chunk to total output
        if on_output is not None:  # If on_output is provided...
            on_output(chunk)  # ...apply it to each chunk

    if timeout_value > 0:
        try:
            p.wait(timeout=timeout_value)
        except subprocess.TimeoutExpired:
            p.kill()
    else:
        p.wait()

    output += p.stdout.read().decode(errors="ignore")

    end = time.perf_counter()
    return CliResult(p.returncode, output, end - start)


def _run_command(
    cmd: str, result_queue: Queue, stream_to_stdout: bool, error_flag: List[bool]
):
    process = subprocess.Popen(
        cmd,
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    # output = ""
    # for line in iter(lambda: process.stdout.read(1), b""):
    #     # if stream_to_stdout:
    #     # sys.stdout.write(line.decode("utf-8", errors="ignore"))
    #     # print(line.decode("utf-8", errors="ignore"), end="")
    #     output += str(line)

    output = process.stdout.read().decode(errors="ignore")

    process.wait()
    result_queue.put(CliResult(process.returncode, output))

    if process.returncode != 0:
        error_flag[0] = True


def run_cli_parallel(
    cmds: List[str], timeout_value=-1, stream_to_stdout=False
) -> List[CliResult]:
    results = []
    result_queue = Queue()
    error_flag = [False]

    threads = []
    for cmd in cmds:
        thread = Thread(
            target=_run_command, args=(cmd, result_queue, stream_to_stdout, error_flag)
        )
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join(timeout=timeout_value)

    # If any process finished with a non-zero code, terminate the remaining processes
    if any(error_flag):
        raise RuntimeError("One or more processes  finished with a non-zero exit code.")

    # Retrieve results from the queue
    while len(results) < len(cmds):
        out = result_queue.get()
        if out is not None:
            results.append(out)

    return results



from typing import List

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution


class EncoderAppleHEVC(Encoder):
    def get_codec(self) -> Codec:
        return Codec.h265

    def get_pretty_name(self) -> str:
        return "APPLE_H265"

    def get_encode_commands(self) -> List[str]:
        vec = [
            self.get_ffmpeg_pipe_command(),
            "|",
            get_binary("ffmpeg"),
            " -hide_banner -y -i -",
            f"-c:v hevc_videotoolbox",
            f"-allow_sw false",
            f"-g {self.keyint}",
        ]

        match self.rate_distribution:
            case EncoderRateDistribution.CQ:
                # the framework has assumptions that crf is 0-63; where 0 is the highest quality,
                # but the videotoolbox api is 100-1 where 100 is the highest quality
                # thats why we need to transform the number
                crf = self.crf
                # scale 0-63 to 1-100
                crf = crf * 100 / 63
                # invert the number
                if crf == 0:
                    crf = 1
                else:
                    crf = 100 - crf
                crf = int(crf)
                vec.append(f"-q:v {crf}")
            case EncoderRateDistribution.VBR:
                vec.append(f"-b:v {int(self.bitrate)}k")
            case _:
                raise NotImplementedError

        if self.bit_override == 10:
            vec.append(f"-profile main10")
        else:
            vec.append(f"-profile main")

        if self.hdr:
            vec.append(f"-color_primaries {self.color_primaries}")
            vec.append(f"-color_trc {self.transfer_characteristics}")
            vec.append(f"-colorspace {self.matrix_coefficients}")
            vec.append(f"-master_display {self.svt_master_display}")
            vec.append(
                f"-max_cll {self.maximum_content_light_level},{self.maximum_frame_average_light_level}"
            )

        vec.append(f'"{self.output_path}"')

        return [" ".join(vec)]

    def get_chunk_file_extension(self) -> str:
        return ".mkv"

    def get_version(self) -> str:
        return "N/A"



import asyncio
import hashlib
import os
import random
import sys

from tqdm import tqdm

from alabamaEncode.cli.cli_setup.paths import parse_paths
from alabamaEncode.cli.cli_setup.ratecontrol import parse_rd
from alabamaEncode.cli.cli_setup.res_preset import parse_resolution_presets
from alabamaEncode.cli.cli_setup.video_filters import parse_video_filters
from alabamaEncode.conent_analysis.opinionated_vmaf import (
    get_vmaf_list,
)
from alabamaEncode.conent_analysis.pipelines import (
    run_sequence_pipeline,
    get_refine_steps,
)
from alabamaEncode.core.chunk_job import ChunkEncoder
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.extras.vmaf_plot import plot_vmaf
from alabamaEncode.core.extras.ws_update import WebsiteUpdate
from alabamaEncode.core.final_touches import (
    print_stats,
    generate_previews,
    create_torrent_file,
)
from alabamaEncode.metrics.impl.vmaf import download_vmaf_models
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.parallel_execution.celery_app import app
from alabamaEncode.parallel_execution.execute_commands import execute_commands
from alabamaEncode.scene.annel import annealing
from alabamaEncode.scene.concat import VideoConcatenator
from alabamaEncode.scene.scene_detection import scene_detect
from alabamaEncode.scene.sequence import ChunkSequence


class AlabamaEncodingJob:
    def __init__(self, ctx):
        self.ctx = ctx
        self.websiteUpdate = WebsiteUpdate(ctx)
        self.finished_callback = None

        self.serialise_folder = self.get_serialise_folder()

        sha1 = hashlib.sha1()
        sha1.update(str.encode(self.ctx.get_title()))
        hash_as_hex = sha1.hexdigest()
        title_hash = int(hash_as_hex, 16) % sys.maxsize
        self.serialise_file = os.path.join(
            self.serialise_folder,
            f"{title_hash}" f".json",
        )

    @staticmethod
    def get_serialise_folder():
        serialise_folder = os.path.expanduser("~/.alabamaEncoder/jobs")
        if not os.path.exists(serialise_folder):
            os.makedirs(serialise_folder)
        return serialise_folder

    @staticmethod
    def get_saved_serialised_jobs():
        for file in os.listdir(AlabamaEncodingJob.get_serialise_folder()):
            with open(
                os.path.join(AlabamaEncodingJob.get_serialise_folder(), file)
            ) as f:
                yield f.read()

    def save(self):
        with open(self.serialise_file, "w") as f:
            f.write(self.ctx.to_json())

    def delete(self):
        if os.path.exists(self.serialise_file):
            os.remove(self.serialise_file)

    @staticmethod
    def load_from_file(json_job):
        ctx = AlabamaContext().from_json(json_job)
        creation_pipeline = [
            parse_paths,
            parse_rd,
            parse_resolution_presets,
            parse_video_filters,
        ]
        for pipeline_item in creation_pipeline:
            ctx = pipeline_item(ctx)

        return AlabamaEncodingJob(ctx)

    def encode_finished(self):
        if self.ctx.multi_res_pipeline:
            return os.path.exists(self.ctx.output_file)
            # TODO: do a paranoid check for all files
        else:
            return os.path.exists(self.ctx.output_file)

    async def run_pipeline(self):
        self.save()

        if self.ctx.use_celery:
            print("Using celery")
            import socket

            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            try:
                # doesn't even have to be reachable
                s.connect(("10.255.255.255", 1))
                host_address = s.getsockname()[0]
            finally:
                s.close()
            print(f"Got lan ip: {host_address}")

            num_workers = app.control.inspect().active_queues()
            if num_workers is None:
                print("No workers detected, please start some")
                quit()
            print(f"Number of available workers: {len(num_workers)}")
        else:
            print(
                f"Using {self.ctx.prototype_encoder.get_pretty_name()} version: {self.ctx.prototype_encoder.get_version()}"
            )

        constant_updates = asyncio.create_task(self.websiteUpdate.constant_updates())
        self.websiteUpdate.update_current_step_name("Running scene detection")

        sequence = self.prepare_sequence()

        if not self.encode_finished():
            self.websiteUpdate.update_proc_done(10)
            self.websiteUpdate.update_current_step_name("Analyzing content")
            download_vmaf_models()
            await run_sequence_pipeline(self.ctx, sequence)
            self.ctx.get_kv().set_global("quiet_analyzing_content_logs", True)

            self.websiteUpdate.update_proc_done(20)
            self.websiteUpdate.update_current_step_name("Encoding scenes")

            iter_counter = 0
            if self.ctx.dry_run:
                iter_counter = 2

            while sequence.sequence_integrity_check(kv=self.ctx.get_kv()):
                iter_counter += 1
                if iter_counter > 3:
                    print("Integrity check failed 3 times, aborting")
                    quit()

                ctx = self.ctx
                frames_encoded_so_far = 0
                size_kb_so_far = 0

                command_objects = []

                def is_chunk_done(_chunk):
                    if ctx.multi_res_pipeline:
                        enc = ctx.get_encoder()
                        vmafs = get_vmaf_list(enc.get_codec())

                        for vmaf in vmafs:
                            output_path = (
                                f"{ctx.temp_folder}/"
                                f"{_chunk.chunk_index}_{vmaf}.{enc.get_chunk_file_extension()}"
                            )
                            if not os.path.exists(output_path):
                                return False

                        return True
                    else:
                        return _chunk.is_done(kv=ctx.get_kv())

                for chunk in sequence.chunks:
                    if not is_chunk_done(chunk):
                        command_objects.append(ChunkEncoder(ctx, chunk))
                    else:
                        frames_encoded_so_far += chunk.get_frame_count()
                        size_kb_so_far += chunk.get_filesize() / 1000

                threads = os.cpu_count()
                if len(command_objects) < threads:
                    ctx.prototype_encoder.threads = int(threads / len(command_objects))

                # order chunks based on order
                if ctx.chunk_order == "random":
                    random.shuffle(command_objects)
                elif ctx.chunk_order == "length_asc":
                    command_objects.sort(key=lambda x: x.chunk.length)
                elif ctx.chunk_order == "length_desc":
                    command_objects.sort(key=lambda x: x.chunk.length, reverse=True)
                elif ctx.chunk_order == "sequential":
                    pass
                elif ctx.chunk_order == "sequential_reverse":
                    command_objects.reverse()
                elif ctx.chunk_order == "even":
                    command_objects = annealing(command_objects, 1000)
                else:
                    raise ValueError(f"Invalid chunk order: {ctx.chunk_order}")

                if ctx.throughput_scaling:
                    # make the chunk length distribution homogenous
                    command_objects = annealing(command_objects, 1000)

                print(
                    f"Starting encoding of {len(command_objects)} out of {len(sequence.chunks)} scenes"
                )

                def update_proc_done(num_finished_scenes):
                    # map 20 to 95% as the space where the scenes are encoded
                    already_done = len(sequence.chunks) - len(command_objects)
                    # self.websiteUpdate.update_proc_done(
                    #     20
                    #     + (already_done + num_finished_scenes)
                    #     / len(sequence.chunks)
                    #     * 75
                    # )

                    self.websiteUpdate.update_proc_done(
                        ((already_done + num_finished_scenes) / len(sequence.chunks))
                        * 100
                    )

                kv = self.ctx.get_kv()
                saved_progress = kv.get_global("pbar_progress")

                pbar = tqdm(
                    total=sum([c.chunk.length for c in command_objects]),
                    desc="Encoding",
                    unit="frame",
                    dynamic_ncols=True,
                    unit_scale=True,
                    smoothing=0.2,
                    initial=saved_progress if saved_progress is not None else 0,
                )

                saved_total = kv.get_global("pbar_total")
                if saved_total is not None:
                    pbar.total = saved_total

                saved_estimation = kv.get_global("pbar_estimation")
                if saved_estimation is not None:
                    pbar.set_postfix(estimation=saved_estimation)

                pbar.refresh()

                if len(command_objects) == 0:
                    print("Nothing to encode, skipping")
                else:
                    try:
                        await asyncio.create_task(
                            execute_commands(
                                ctx.use_celery,
                                command_objects,
                                ctx.multiprocess_workers,
                                pin_to_cores=ctx.pin_to_cores,
                                finished_scene_callback=update_proc_done,
                                size_estimate_data=(
                                    frames_encoded_so_far,
                                    size_kb_so_far,
                                ),
                                throughput_scaling=ctx.throughput_scaling,
                                pbar=pbar,
                            )
                        )
                    except (KeyboardInterrupt, asyncio.exceptions.CancelledError):
                        print("Keyboard interrupt, stopping")
                        # kill all async tasks
                        pbar.close()
                        for task in asyncio.all_tasks():
                            task.cancel()

                        # save pbar progress in kv
                        kv = self.ctx.get_kv()
                        kv.set_global("pbar_progress", pbar.n)
                        if kv.get_global("pbar_total") is None:
                            kv.set_global("pbar_total", pbar.total)
                        kv.set_global("pbar_estimation", pbar.postfix.get("estimation"))
                        quit()

                refine_steps = get_refine_steps(ctx)
                for step in refine_steps:
                    step(ctx, sequence)

            if not self.ctx.multi_res_pipeline:
                self.websiteUpdate.update_proc_done(95)
                self.websiteUpdate.update_current_step_name("Concatenating scenes")

                try:
                    VideoConcatenator(
                        output=self.ctx.output_file,
                        file_with_audio=self.ctx.input_file,
                        audio_param_override=self.ctx.audio_params,
                        start_offset=self.ctx.start_offset,
                        end_offset=self.ctx.end_offset,
                        title=self.ctx.get_title(),
                        encoder_name=self.ctx.encoder_name,
                        mux_audio=self.ctx.encode_audio,
                        subs_file=[self.ctx.sub_file],
                        temp_dir=self.ctx.temp_folder,
                    ).find_files_in_dir(
                        folder_path=os.path.join(self.ctx.temp_folder, "chunks"),
                        extension=self.ctx.get_encoder().get_chunk_file_extension(),
                    ).concat_videos()
                except Exception as e:
                    print("Concat failed 😷")
                    raise e
            constant_updates.cancel()
        else:
            print("Output file already exists")

        self.websiteUpdate.update_proc_done(99)
        self.websiteUpdate.update_current_step_name("Final touches")

        if self.ctx.generate_stats:
            print_stats(
                output_folder=self.ctx.output_folder,
                output=self.ctx.output_file,
                # input_file=self.ctx.raw_input_file,
                # grain_synth=self.ctx.prototype_encoder.grain_synth,
                title=self.ctx.get_title(),
                # cut_intro=(True if self.ctx.start_offset > 0 else False),
                # cut_credits=(True if self.ctx.end_offset > 0 else False),
                # croped=(True if self.ctx.crop_string != "" else False),
                # scaled=(True if self.ctx.scale_string != "" else False),
                # tonemaped=(
                #     True
                #     if not self.ctx.prototype_encoder.hdr
                #     and Ffmpeg.is_hdr(PathAlabama(self.ctx.input_file))
                #     else False
                # ),
            )
            if (
                self.ctx.calc_final_vmaf
                and self.ctx.get_metric_target()[0] == Metric.VMAF
            ):
                plot_vmaf(self.ctx, sequence)

        if self.ctx.generate_previews:
            generate_previews(
                input_file=self.ctx.output_file, output_folder=self.ctx.output_folder
            )
            create_torrent_file(
                video=self.ctx.output_file,
                encoder_name=self.ctx.encoder_name,
                output_folder=self.ctx.output_folder,
            )

        # TODO cleanup right in multiencode
        print("Cleaning up temp folder 🥺")
        for root, dirs, files in os.walk(self.ctx.temp_folder):
            # remove all folders that contain 'rate_probes'
            for name in dirs:
                if "rate_probes" in name:
                    # remove {rate probe folder}/*.ivf
                    for root2, dirs2, files2 in os.walk(self.ctx.temp_folder + name):
                        for name2 in files2:
                            if name2.endswith(".ivf"):
                                os.remove(self.ctx.temp_folder + name + "/" + name2)
            # remove all *.stat files in tempfolder
            for name in files:
                if name.endswith(".stat"):
                    # try to remove
                    os.remove(self.ctx.temp_folder + name)
        # clean empty folders in the temp folder
        for root, dirs, files in os.walk(self.ctx.temp_folder):
            for name in dirs:
                if len(os.listdir(os.path.join(root, name))) == 0:
                    os.rmdir(os.path.join(root, name))

        self.websiteUpdate.update_proc_done(100)
        self.websiteUpdate.update_current_step_name("Done")
        if self.finished_callback is not None:
            self.finished_callback()

        self.delete()

    def prepare_sequence(self):
        sequence: ChunkSequence = scene_detect(
            input_file=self.ctx.input_file,
            cache_file_path=self.ctx.temp_folder + "scene_cache.json",
            max_scene_length=self.ctx.max_scene_length,
            start_offset=self.ctx.start_offset,
            end_offset=self.ctx.end_offset,
            override_bad_wrong_cache_path=self.ctx.override_scenecache_path_check,
            static_length=self.ctx.statically_sized_scenes,
            static_length_size=self.ctx.max_scene_length,
            scene_merge=self.ctx.scene_merge,
        )
        sequence.setup_paths(
            temp_folder=self.ctx.temp_folder,
            extension=self.ctx.get_encoder().get_chunk_file_extension(),
        )
        self.ctx.total_chunks = len(sequence.chunks)
        return sequence



import os
import pickle
import time

from alabamaEncode.conent_analysis.chunk.chunk_analyse_step import (
    ChunkAnalyzePipelineItem,
)
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.metrics.impl.ssim import get_video_ssim
from alabamaEncode.scene.chunk import ChunkObject


class VbrPerChunkOptimised(ChunkAnalyzePipelineItem):
    def run(self, ctx: AlabamaContext, chunk: ChunkObject, enc: Encoder) -> Encoder:
        enc.rate_distribution = EncoderRateDistribution.VBR
        enc.bitrate = get_ideal_bitrate(chunk, ctx)
        enc.passes = 3
        return enc


def get_ideal_bitrate(
    chunk,
    config,
    convex_speed=10,
    show_rate_calc_log=False,
    clamp_complexity=True,
) -> int:
    """
    Gets the ideal bitrate for a chunk
    """

    rate_search_start = time.time()

    probe_file_base = config.get_probe_file_base(chunk.chunk_path)
    cache_filename = f"{probe_file_base}complexity.speed{convex_speed}.pt"

    ideal_rate = None

    # check if we have already done this
    if os.path.exists(cache_filename):
        try:
            ideal_rate = pickle.load(open(cache_filename, "rb"))
        except:
            print("Error loading complexity cache file, recalculating")

    if not ideal_rate:
        encoder = config.get_encoder()

        test_probe_path = (
            f"{probe_file_base}complexity.probe{encoder.get_chunk_file_extension()}"
        )

        encoder.speed = convex_speed
        encoder.passes = 1
        encoder.temp_folder = config.temp_folder
        encoder.chunk = chunk
        encoder.grain_synth = 0
        encoder.current_scene_index = chunk.chunk_index
        encoder.output_path = test_probe_path
        encoder.threads = 1
        encoder.video_filters = config.prototype_encoder.video_filters
        encoder.rate_distribution = EncoderRateDistribution.VBR

        encoder.run(override_if_exists=False)

        try:
            (ssim, ssim_db) = get_video_ssim(
                test_probe_path,
                chunk,
                get_db=True,
                video_filters=config.prototype_encoder.video_filters,
            )
        except Exception as e:
            print(f"Error calculating ssim for complexity rate estimation: {e}")
            # this happens when the scene is fully black, the best solution here is just setting the complexity to 0,
            # and since its all black anyway it won't matter
            ssim_db = config.ssim_db_target

        # Calculate the ratio between the target ssim dB and the current ssim dB
        ratio = 10 ** ((config.ssim_db_target - ssim_db) / 10)

        # Clamp the ratio to the complexity clamp
        if clamp_complexity:
            ratio = max(
                min(ratio, 1 + config.bitrate_overshoot), 1 - config.bitrate_undershoot
            )

        # Interpolate the ideal rate using the ratio
        ideal_rate = config.prototype_encoder.bitrate * ratio
        ideal_rate = int(ideal_rate)

        if show_rate_calc_log:
            print(
                f"{chunk.log_prefix()}===============\n"
                f"{chunk.log_prefix()} encode rate: {config.prototype_encoder.bitrate}k/s\n"
                f"{chunk.log_prefix()} ssim dB when using target bitrate: {ssim_db} (wanted: {config.ssim_db_target})\n"
                f"{chunk.log_prefix()} ratio = 10 ** (dB_target - dB) / 10 = {ratio}\n"
                f"{chunk.log_prefix()} ideal rate: max(min(encode_rate * ratio, upper_clamp), bottom_clamp) = {ideal_rate:.2f}k/s\n"
                f"{chunk.log_prefix()}==============="
            )

        try:
            pickle.dump(ideal_rate, open(cache_filename, "wb"))
        except:
            pass

    if ideal_rate == -1:
        raise Exception("ideal_rate is -1")

    config.log(
        f"{chunk.log_prefix()}rate search took: {int(time.time() - rate_search_start)}s, ideal bitrate: {ideal_rate}"
    )

    return int(ideal_rate)



from abc import ABC

from alabamaEncode.metrics.comparison_display import ComparisonDisplayResolution


class MetricOptions(ABC):
    ref: ComparisonDisplayResolution = None
    denoise_reference = False
    video_filters = ""
    threads = 1

    def __init__(
        self,
        **kwargs,
    ):
        if kwargs is not None:
            self.__dict__.update(kwargs)



import os
import random
import struct

from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.scene.chunk import ChunkObject

# big shoutout to @master_of_zen for this

fields = [
    "frame",
    "weight",
    "intra_error",
    "frame_avg_wavelet_energy",
    "coded_error",
    "sr_coded_error",
    "tr_coded_error",
    "pcnt_inter",
    "pcnt_motion",
    "pcnt_second_ref",
    "pcnt_third_ref",
    "pcnt_neutral",
    "intra_skip_pct",
    "inactive_zone_rows",
    "inactive_zone_cols",
    "MVr",
    "mvr_abs",
    "MVc",
    "mvc_abs",
    "MVrv",
    "MVcv",
    "mv_in_out_count",
    "new_mv_count",
    "duration",
    "count",
    "raw_error_stdev",
]


def aom_extract_firstpass_data(chunk: ChunkObject, vf=""):
    pass_file_path = f"/tmp/{random.randint(0, 100000)}.pass"

    run_cli(
        f"{chunk.create_chunk_ffmpeg_pipe_command(video_filters=vf)} | {get_binary('aomenc')} - "
        f"--ivf --fpf={pass_file_path} --threads=8 --passes=2 --pass=1 --auto-alt-ref=1 --lag-in-frames=25 "
        f"-o {os.devnull}"
    ).verify()

    dict_list = []

    with open(pass_file_path, "rb") as file:
        frame_buf = file.read(208)
        while len(frame_buf) > 0 and len(frame_buf) % 208 == 0:
            stats = struct.unpack("d" * 26, frame_buf)
            p = dict(zip(fields, stats))
            dict_list.append(p)
            frame_buf = file.read(208)

    os.remove(pass_file_path)

    # print(json.dumps(dict_list, indent=4))
    return dict_list



from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama


def validate_input(ctx):
    try:
        tracks = Ffmpeg.get_tracks(PathAlabama(ctx.raw_input_file))
        video_track = None
        for track in tracks:
            if track["codec_type"] == "video":
                video_track = track
                break

        if video_track is None:
            print("Cant find video track in input file")
            quit()

        if video_track["codec_name"] == "vc1":
            print(
                "Input file is VC1, VC1 ffmpeg seeking is broken, you'll need encode a lossless proxy to fix this"
            )
            quit()

        hdr = True
        if 'color_transfer' not in video_track:
            hdr = False
        else:
            if (
                "bt709" in video_track["color_transfer"]
                or "unknown" in video_track["color_transfer"]
            ):
                hdr = False

        dem, num = video_track["avg_frame_rate"].split("/")
        fps_rounded = "{:.2f}".format(float(dem) / float(num))
        vid_bitrate = ctx.get_kv().get_global("video_bitrate_formatted")
        if vid_bitrate is None:
            print("Getting source bitrate...")
            vid_bitrate = "{:.2f}".format(
                (
                    Ffmpeg.get_source_bitrates(
                        PathAlabama(ctx.raw_input_file), calculate_audio=False
                    )[0]
                    / 1000
                )
            )
            ctx.get_kv().set_global("video_bitrate_formatted", vid_bitrate)
        codec = video_track['codec_name'].upper()
        print(
            f"Input Video: {video_track['width']}x{video_track['height']} @ {fps_rounded} fps, {video_track['pix_fmt'].upper()}, {'HDR' if hdr else 'SDR'}, {vid_bitrate} Kb/s, {codec}"
        )

    except Exception as e:
        print(f'Input file parsing failed: "{e}", is it a valid video file?')
        quit()
    return ctx



import os

from alabamaEncode.conent_analysis.chunk.analyze_steps.target_vmaf import TargetVmaf
from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.util.bin_utils import register_bin
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.metrics.comparison_display import ComparisonDisplayResolution
from alabamaEncode.metrics.impl.vmaf import VmafOptions
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.scene_detection import scene_detect
from alabamaEncode.scene.sequence import ChunkSequence

if __name__ == "__main__":
    env = "./seq_convexhull"
    env = os.path.abspath(env)
    if not os.path.exists(env):
        os.mkdir(env)

    scenes: ChunkSequence = scene_detect(
        input_file="/home/kokoniara/ep3_halo_test.mkv",
        max_scene_length=10,
        cache_file_path=f"{env}/scenes_skinny.pt",
    )

    scenes.setup_paths(env, ".mkv")

    register_bin("SvtAv1EncApp", "/home/kokoniara/.local/opt/SvtAv1EncApp")
    ref_display = ComparisonDisplayResolution.FHD
    resolutions = [
        "1920:-2",
        "1366:-2",
        "1280:-2",
        "960:-2",
        # "854:-2",
        # "768:-2",
        # "640:-2",
        # "480:-2",
    ]
    vmaf_targets = [45, 55, 62, 68, 81, 87, 90, 93, 95, 97]
    ctx = AlabamaContext()
    ctx.vmaf_reference_display = "FHD"
    ctx.temp_folder = env

    enc = EncoderX264()
    enc.speed = 2

    for scene in scenes.chunks:
        for res in resolutions:
            ctx.vmaf = 95
            enc.video_filters = f"crop=3840:1920:0:120,scale={res}"
            scene.chunk_path = os.path.join(env, f"{res.split(':')[0]}p_{ctx.vmaf}.mkv")
            enc.chunk = scene
            ctx.probe_speed = enc.speed
            tar = TargetVmaf()
            tar.run(chunk=scene, ctx=ctx, enc=enc)
            print(f"res: {res}, target 95, computed crf: {enc.crf}")

            stats = enc.run(
                metric_to_calculate=Metric.VMAF,
                metric_params=VmafOptions(ref=ref_display),
            )
            print(
                f"target 95, res: {res}, final vmaf: {stats.vmaf} crf: {enc.crf} bitrate: {stats.bitrate}"
            )



import json
import os

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.scene.sequence import ChunkSequence


def scrape_hdr_metadata(ctx: AlabamaContext, sequence: ChunkSequence):
    """
    Scrapes HDR metadata from the input file
    """
    if ctx.prototype_encoder.hdr and (
        isinstance(ctx.prototype_encoder, EncoderSvt)
        or isinstance(ctx.prototype_encoder, EncoderX264)
    ):
        if not Ffmpeg.is_hdr(PathAlabama(ctx.raw_input_file)):
            print("Input file is not HDR, disabling HDR mode")
            ctx.prototype_encoder.hdr = False
            return ctx

        cache_path = f"{ctx.temp_folder}hdr.cache"

        if not os.path.exists(cache_path):
            print("Running auto HDR10")
            obj = Ffmpeg.get_first_frame_data(PathAlabama(ctx.raw_input_file))
            color_space = obj["color_space"]
            if "bt2020nc" in color_space:
                color_space = "bt2020-ncl"

            if ctx.prototype_encoder.matrix_coefficients == "bt709":
                ctx.prototype_encoder.matrix_coefficients = color_space
                print(
                    f"Setting color space to {ctx.prototype_encoder.matrix_coefficients}"
                )

            if ctx.prototype_encoder.color_primaries == "bt709":
                ctx.prototype_encoder.color_primaries = obj["color_primaries"]
                print(
                    f"Setting color primaries to {ctx.prototype_encoder.color_primaries}"
                )

            if ctx.prototype_encoder.transfer_characteristics == "bt709":
                ctx.prototype_encoder.transfer_characteristics = obj["color_transfer"]
                print(
                    f"Setting transfer characteristics to {ctx.prototype_encoder.transfer_characteristics}"
                )

            ctx.prototype_encoder.chroma_sample_position = obj["chroma_location"]
            print(
                f"Setting chroma sample position to {ctx.prototype_encoder.chroma_sample_position}"
            )

            for side_data in obj["side_data_list"]:
                if side_data["side_data_type"] == "Content light level metadata":
                    if ctx.prototype_encoder.maximum_content_light_level == "":
                        ctx.prototype_encoder.maximum_content_light_level = side_data[
                            "max_content"
                        ]
                    if ctx.prototype_encoder.maximum_frame_average_light_level == "":
                        ctx.prototype_encoder.maximum_frame_average_light_level = (
                            side_data["max_average"]
                        )
                    print(
                        f"Setting max content light level to {ctx.prototype_encoder.maximum_content_light_level}"
                    )
                    print(
                        f"Setting max frame average light level to "
                        f"{ctx.prototype_encoder.maximum_frame_average_light_level}"
                    )
                if side_data["side_data_type"] == "Mastering display metadata":

                    def split_and_divide(spltting) -> float:
                        spl = spltting.split("/")
                        return int(spl[0]) / int(spl[1])

                    red_x = split_and_divide(side_data["red_x"])
                    red_y = split_and_divide(side_data["red_y"])
                    green_x = split_and_divide(side_data["green_x"])
                    green_y = split_and_divide(side_data["green_y"])
                    blue_x = split_and_divide(side_data["blue_x"])
                    blue_y = split_and_divide(side_data["blue_y"])
                    white_point_x = split_and_divide(side_data["white_point_x"])
                    white_point_y = split_and_divide(side_data["white_point_y"])
                    min_luminance = split_and_divide(side_data["min_luminance"])
                    max_luminance = split_and_divide(side_data["max_luminance"])
                    # G(x,y)B(x,y)R(x,y)WP(x,y)L(max,min)
                    ctx.prototype_encoder.svt_master_display = (
                        f"G({green_x},{green_y})B({blue_x},{blue_y})R({red_x},{red_y})"
                        f"WP({white_point_x},{white_point_y})L({max_luminance},{min_luminance})"
                    )

                    print(
                        f"Setting svt master display to {ctx.prototype_encoder.svt_master_display}"
                    )

            cache_obj = {
                "matrix_coefficients": ctx.prototype_encoder.matrix_coefficients,
                "color_primaries": ctx.prototype_encoder.color_primaries,
                "transfer_characteristics": ctx.prototype_encoder.transfer_characteristics,
                "chroma_sample_position": ctx.prototype_encoder.chroma_sample_position,
                "maximum_content_light_level": ctx.prototype_encoder.maximum_content_light_level,
                "maximum_frame_average_light_level": ctx.prototype_encoder.maximum_frame_average_light_level,
                "svt_master_display": ctx.prototype_encoder.svt_master_display,
            }

            # save as json
            with open(cache_path, "w") as f:
                f.write(json.dumps(cache_obj))
        else:
            print("Loading HDR10 metadata from cache")
            with open(cache_path) as f:
                cache_obj = json.loads(f.read())
            ctx.prototype_encoder.matrix_coefficients = cache_obj["matrix_coefficients"]
            ctx.prototype_encoder.color_primaries = cache_obj["color_primaries"]
            ctx.prototype_encoder.transfer_characteristics = cache_obj[
                "transfer_characteristics"
            ]
            ctx.prototype_encoder.chroma_sample_position = cache_obj[
                "chroma_sample_position"
            ]
            ctx.prototype_encoder.maximum_content_light_level = cache_obj[
                "maximum_content_light_level"
            ]
            ctx.prototype_encoder.maximum_frame_average_light_level = cache_obj[
                "maximum_frame_average_light_level"
            ]
            ctx.prototype_encoder.svt_master_display = cache_obj["svt_master_display"]

    return ctx


if __name__ == "__main__":
    ctx = AlabamaContext()
    ctx.prototype_encoder.hdr = True
    # ctx.raw_input_file = "/mnt/data/downloads/Halo.S01.2160p.UHD.BluRay.Remux.HDR.DV.HEVC.Atmos-PmP/Halo.S01E04.Homecoming.2160p.UHD.BluRay.Remux.HDR.DV.HEVC.Atmos-PmP.mkv"
    ctx.raw_input_file = "/mnt/data/downloads/Silo.S01E02.HDR.2160p.WEB.H265-GGEZ[rarbg]/silo.s01e02.hdr.2160p.web.h265-ggez.mkv"
    print(f"Extracting HDR10 metadata 🤩😳 for {os.path.basename(ctx.raw_input_file)}")
    ctx = scrape_hdr_metadata(ctx)

    print("matrix_coefficients", ctx.prototype_encoder.matrix_coefficients)
    print("color_primaries", ctx.prototype_encoder.color_primaries)
    print("transfer_characteristics", ctx.prototype_encoder.transfer_characteristics)
    print("chroma_sample_position", ctx.prototype_encoder.chroma_sample_position)
    print("max_content", ctx.prototype_encoder.maximum_content_light_level)
    print("max_average", ctx.prototype_encoder.maximum_frame_average_light_level)
    print("svt master display: ", ctx.prototype_encoder.svt_master_display)



import os
import re

from alabamaEncode.core.context import AlabamaContext


def parse_movie_title(title):
    match = re.match(r"^(.*?) \((\d{4})\)$", title)
    if match:
        return match.group(1), match.group(2)
    return None


def parse_tv_title(title):
    """
    Parses a tv title and returns the title, year, season and episode
    """
    match = re.match(r"^(.*?) \((\d{4})\) S(\d{2})E(\d{2})$", title)
    if match:
        return match.group(1), match.group(2), match.group(3), match.group(4)
    return None


def is_title_movie(title):
    """
    Returns true if the title is a movie
    """
    return title[-1] == ")"


def is_valid_title(title):
    """
    Returns true if the title is a movie or tv show
    """
    return is_title_movie(title) or re.match(
        r"^(.*?) \((\d{4})\) S(\d{2})E(\d{2})$", title
    )


def auto_output_paths(ctx):
    """
    If output file is set to "auto" compute the output file name from title
    """
    if ctx.output_file == "auto":
        if ctx.title == "":
            print(
                '`--title "Title (year) [S--E--|Nothing for movies]"` must be set if output file is set to auto'
            )
            quit()
        if not is_valid_title(ctx.title):
            print(
                f"Title {ctx.title} is not valid, must be in the format `Title (year) [SExx|Nothing for movies]`"
            )
            quit()

        is_movie = is_title_movie(title=ctx.title)
        # ~/showsEncode/Title/sx/ex/Title.YEAR.S0xE0x.{ctx.encoder_name}.webm
        # ~/movieEncode/Title/Title.YEAR.{ctx.encoder_name}.webm

        hdr_bit = ".HDR10" if ctx.prototype_encoder.hdr else ""
        res_preset = "." + ctx.resolution_preset if ctx.resolution_preset != "" else ""

        if is_movie:
            title, year = parse_movie_title(ctx.title)
            ctx.output_file = (
                f"~/movieEncode/{title} ({year})/"
                f"{title.replace(' ', '.')}.{year}{res_preset}.AV1.OPUS{hdr_bit}.{ctx.encoder_name}.webm"
            )
        else:
            title, year, season, episode = parse_tv_title(ctx.title)
            ctx.output_file = (
                f"~/showsEncode/{title} ({year})/s{int(season)}/e{int(episode)}/"
                f"{title.replace(' ', '.')}.{year}.S{season}E{episode}{res_preset}"
                f".AV1.OPUS{hdr_bit}.{ctx.encoder_name}.webm"
            )

        ctx.output_file = os.path.expanduser(ctx.output_file)

    return ctx


def movie_parser_test():
    title = "The Matrix (1999)"
    title, year = parse_movie_title(title)
    assert title == "The Matrix"
    assert year == "1999"

    title = "Movie (2020)"
    title, year = parse_movie_title(title)
    assert title == "Movie"
    assert year == "2020"


def tv_parser_test():
    title = "The Matrix (1999) S01E01"
    title, year, season, episode = parse_tv_title(title)
    assert title == "The Matrix"
    assert year == "1999"
    assert season == "01"
    assert episode == "01"


def title_validator_test():
    title = "The Matrix (1999)"
    assert is_valid_title(title)

    title = "The Matrix (1999) S01E01"
    assert is_valid_title(title)

    title = "The matrix 1999 S01E01"
    assert not is_valid_title(title)

    title = "The Matrix S01E01"
    assert not is_valid_title(title)

    title = "The Matrix () S01E01"
    assert not is_valid_title(title)


def test_auto_paths():
    ctx = AlabamaContext()
    ctx.encoder_name = "SouAV1R"
    home = os.path.expanduser("~")

    ctx.output_file = "auto"
    ctx.title = "The Matrix (1999)"
    ctx = auto_output_paths(ctx)
    print(ctx.output_file)
    assert (
        ctx.output_file
        == f"{home}/movieEncode/The Matrix (1999)/The.Matrix.1999.AV1.OPUS.SouAV1R.webm"
    )

    ctx.output_file = "auto"
    ctx.title = "The Matrix (1999) S01E01"
    ctx = auto_output_paths(ctx)

    print(ctx.output_file)
    assert (
        ctx.output_file
        == f"{home}/showsEncode/The Matrix (1999)/s1/e1/The.Matrix.1999.S01E01.AV1.OPUS.SouAV1R.webm"
    )


if __name__ == "__main__":
    movie_parser_test()
    tv_parser_test()
    title_validator_test()
    test_auto_paths()



import asyncio
import os
import sys

from aiohttp.web_app import Application
from aiohttp.web_routedef import post
from aiohttp.web_runner import AppRunner, TCPSite

from alabamaEncode.core.job import AlabamaEncodingJob


async def worker(name, queue):
    while True:
        try:
            job = await queue.get()
            await job.run_pipeline()
            queue.task_done()
        except Exception as e:
            print(f"Worker {name} got exception)")
            raise e


async def main():
    worker_count = 1
    queue = asyncio.Queue()
    workers = []
    for i in range(worker_count):
        w = asyncio.create_task(worker(f"worker-{i}", queue))
        workers.append(w)

    # first: queue up old jobs
    # read json jobs from ~/.alabamaEncoder/jobs/*.json, serialize into contexts, build jobs and queue up
    for serialised_job in AlabamaEncodingJob.get_saved_serialised_jobs():
        job = AlabamaEncodingJob.load_from_file(serialised_job)
        if queue.qsize() > 0:
            job.update_current_step_name("Queued")
        print(f"Found and loaded encoding job for {job.ctx.get_title()}")
        queue.put_nowait(job)

    # async version of the server:
    import aiohttp

    auth_bearer = os.environ.get("AUTH_BEARER_TOKEN", "")
    if auth_bearer == "":
        print("No AUTH_BEARER_TOKEN set, exiting")
        sys.exit(1)

    async def handle(request):
        if request.headers.get("Authorization") != f"Bearer {auth_bearer}":
            return aiohttp.web.Response(status=401)

        post_data = await request.read()
        try:
            _job = AlabamaEncodingJob.load_from_file(post_data.decode())
        except Exception as e:
            print(f"Failed to load context from json: {e}")
            return aiohttp.web.Response(status=400)
        if queue.qsize() > 0:
            job.update_current_step_name("Queued")
        print(f"Received encoding job for {_job.ctx.get_title()}")
        queue.put_nowait(_job)
        return aiohttp.web.Response(status=200)

    print("Starting server")
    app = Application()
    app.add_routes([post("/jobs", handle)])
    runner = AppRunner(app)
    await runner.setup()
    site = TCPSite(runner, "", 8000)
    await site.start()


if __name__ == "__main__":
    asyncio.run(main())



import concurrent
import copy
import os
import pickle
from concurrent.futures import ThreadPoolExecutor as Executor
from math import log

import keras
import numpy as np
import tensorflow as tf
from keras import layers
from matplotlib import pyplot as plt

from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.metric import Metric

print(tf.__version__)

from alabamaEncode.core.util.bin_utils import register_bin
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.scene.chunk import ChunkObject
from alabamaEncode.scene.sequence import ChunkSequence
from alabamaEncode.scene.scene_detection import scene_detect


def get_complexity(enc: Encoder, c: ChunkObject) -> float:
    _enc = copy.deepcopy(enc)
    _enc.chunk = c
    _enc.speed = 12
    _enc.passes = 1
    _enc.rate_distribution = EncoderRateDistribution.CQ
    _enc.crf = 16
    _enc.threads = 1
    _enc.grain_synth = 0
    _enc.output_path = (
        f"/tmp/{c.chunk_index}_complexity{_enc.get_chunk_file_extension()}"
    )
    stats: EncodeStats = _enc.run()
    formula = log(stats.bitrate)
    # self.config.log(
    #     f"[{c.chunk_index}] complexity: {formula:.2f} in {stats.time_encoding}s"
    # )
    os.remove(_enc.output_path)
    return c.chunk_index, formula


def generate_stat_from_chunk(current_chunk: ChunkObject):
    cache_file_name = os.path.join(env, f"{current_chunk.chunk_index}.pkl")

    if os.path.exists(cache_file_name):
        with open(cache_file_name, "rb") as f:
            return pickle.load(f)

    # current_chunk = test_clips[0]
    motion = Ffmpeg.get_vmaf_motion(current_chunk)
    enc = EncoderSvt()
    enc.chunk = current_chunk
    enc.chunk = current_chunk
    enc.chunk.path = current_chunk.path
    enc.video_filters = current_chunk.video_filters
    complexity = get_complexity(enc, current_chunk)
    enc.speed = 4
    enc.passes = 1
    enc.threads = 2
    enc.temp_folder = env

    local_stats = []
    for crf in range(lower_crf, upper_crf, 2):
        enc.crf = crf
        enc.output_path = os.path.join(
            env,
            (
                f"{os.path.basename(current_chunk.chunk_path)}_"
                f"{crf}{enc.get_chunk_file_extension()}"
            ),
        )
        if os.path.exists(enc.output_path):
            os.remove(enc.output_path)
        stat = enc.run(metric_to_calculate=Metric.VMAF)
        stat.basename = [crf, motion, complexity]
        print(
            f"{current_chunk.log_prefix()}VMAF: {stat.vmaf}, motion: {motion}, complexity: {complexity} crf: {crf}"
        )
        stat.version = current_chunk.chunk_index
        local_stats.append(stat)
        os.remove(enc.output_path)

    vmaf_at_32 = [s.vmaf for s in local_stats if s.basename[0] == 32][0]
    for stat in local_stats:
        stat.basename.append(vmaf_at_32)

    with open(cache_file_name, "wb") as f:
        pickle.dump(local_stats, f)
    if not os.path.exists(cache_file_name):
        raise Exception("Failed to save cache file")

    return local_stats


def get_chunks_of_file(path, video_filters):
    scenes: ChunkSequence = scene_detect(
        input_file=path, cache_file_path=f"{env}/sceneCache_{os.path.basename(path)}.pt"
    )

    for c in scenes.chunks:
        c.video_filters = video_filters

    return scenes.chunks


def get_scenes():
    analyzed_chunks = []

    # test_files = [
    #     "/mnt/data/objective-1-fast/ducks_take_off_1080p50_60f.y4m",
    #     "/mnt/data/objective-1-fast/aspen_1080p_60f.y4m",
    #     "/mnt/data/objective-1-fast/rush_hour_1080p25_60f.y4m",
    #     "/mnt/data/objective-1-fast/touchdown_pass_1080p_60f.y4m",
    #     "/mnt/data/objective-1-fast/KristenAndSara_1280x720_60f.y4m",
    #     "/mnt/data/objective-1-fast/vidyo1_720p_60fps_60f.y4m",
    #     "/mnt/data/objective-1-fast/vidyo4_720p_60fps_60f.y4m",
    #     "/mnt/data/objective-1-fast/gipsrestat720p_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_DrivingPOV_1280x720_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_RollerCoaster_1280x720_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/blue_sky_360p_60f.y4m",
    #     "/mnt/data/objective-1-fast/shields_640x360_60f.y4m",
    #     "/mnt/data/objective-1-fast/speed_bag_640x360_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_Aerial_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_Boat_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_Crosswalk_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_FoodMarket_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_PierSeaside_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_SquareAndTimelapse_1920x1080_60fps_8bit_420_60f.y4m",
    #     "/mnt/data/objective-1-fast/Netflix_TunnelFlag_1920x1080_60fps_8bit_420_60f.y4m",
    # ]
    #
    # for i, f in enumerate(test_files):
    #     analyzed_chunks += get_chunks_of_file(f, "")

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/coraline_sc.mkv", f"{Ffmpeg.get_tonemap_vf()}"
    )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/ep3_halo_test.mkv",
        f"crop=3840:1920:0:120,scale=1920:-2,{Ffmpeg.get_tonemap_vf()}",
    )

    analyzed_chunks += get_chunks_of_file("/mnt/data/liveaction_bright.mkv", "")

    analyzed_chunks += get_chunks_of_file(
        "/mnt/data/liveAction_normal.mp4", "crop=3808:1744:28:208,scale=1920:-2"
    )

    analyzed_chunks += get_chunks_of_file(
        "/home/kokoniara/howlscastle_test.mkv" "crop=1920:1024:0:28"
    )

    for i, c in enumerate(analyzed_chunks):
        c.chunk_path = os.path.join(env, f"{i}.ivf")
        c.chunk_index = i

    return analyzed_chunks


if __name__ == "__main__":
    env = "./crf_vmaf_relation"
    env = os.path.abspath(env)
    if not os.path.exists(env):
        os.mkdir(env)

    analyzed_chunks = get_scenes()

    register_bin(
        "SvtAv1EncApp", "/home/kokoniara/dev/SVT-FASTER/Bin/Release/SvtAv1EncApp"
    )

    lower_crf = 18
    upper_crf = 55

    stats = []

    crfs = range(lower_crf, upper_crf, 2)

    stats_file = os.path.join(env, "stats.pkl")
    if os.path.exists(stats_file):
        with open(stats_file, "rb") as f:
            stats = pickle.load(f)
    else:
        with Executor(max_workers=3) as ex:
            futures = {
                ex.submit(generate_stat_from_chunk, current_chunk): current_chunk
                for current_chunk in analyzed_chunks
            }
            for future in concurrent.futures.as_completed(futures):
                chunk = futures[future]
                try:
                    data = future.result()
                except Exception as exc:
                    print(f"Generated an exception: {exc}")
                else:
                    stats += data

        with open(stats_file, "wb") as f:
            pickle.dump(stats, f)

    stats_by_version = {}
    for stat in stats:
        # if stat version matches a version in the dict, append to that list, else create a new list
        if stat.version in stats_by_version:
            stats_by_version[stat.version].append(stat)
        else:
            stats_by_version[stat.version] = [stat]

    # since crf points are two apart, do interpolation of crf and vmaf and create in-between points
    for version in stats_by_version:
        stats_by_version[version] = sorted(
            stats_by_version[version], key=lambda s: s.basename[0]
        )
        new_stats = []
        for i, stat in enumerate(stats_by_version[version]):
            # first stat is always the same
            if i == 0:
                new_stats.append(stat)
                continue

            # we get the prev stat and then create a interpolated between the curr and last
            prev_stat = stats_by_version[version][i - 1]
            curr_stat = stat
            interpolated_stat = copy.deepcopy(curr_stat)
            interpolated_stat.basename[0] = prev_stat.basename[0] + 1
            interpolated_stat.vmaf = (
                prev_stat.vmaf + curr_stat.vmaf
            ) / 2  # average the vmaf
            new_stats.append(interpolated_stat)
            new_stats.append(curr_stat)

        stats_by_version[version] = new_stats

    train_samples = []  # [[crf at vmaf 96], [motion, complexity]]
    for version in stats_by_version:
        crf_of_vmaf96 = 0
        closest_vmaf = 0
        # search for the crf that is the closest to vmaf 96
        for stat in stats_by_version[version]:
            if abs(stat.vmaf - 96) < abs(closest_vmaf - 96):
                closest_vmaf = stat.vmaf
                crf_of_vmaf96 = stat.basename[0]

        train_samples.append(
            [
                [crf_of_vmaf96],
                [
                    stats_by_version[version][0].basename[1],  # motion
                    stats_by_version[version][0].basename[2],  # complexity
                    # stats_by_version[version][0].basename[3],  # ref
                ],
            ]
        )

    # y is the crf at vmaf 96, x is the motion and complexity
    x = np.array([s[1] for s in train_samples])
    y = np.array([s[0] for s in train_samples])

    print("Size of x: ", x.shape)
    print("Size of y: ", y.shape)

    # (
    #     x_train,
    #     x_test,
    #     y_train,
    #     y_test,
    # ) = train_test_split(x, y, test_size=0.1, random_state=42)
    x_train = x
    y_train = y

    trained_model_path = os.path.join(env, "latest.keras")

    if not os.path.exists(trained_model_path):
        normalisation_layer = layers.Normalization(
            input_shape=[
                x_train.shape[1],
            ]
        )
        normalisation_layer.adapt(x_train)

        # Build a neural network model
        model = tf.keras.Sequential(
            [
                normalisation_layer,
                layers.Dense(64, activation="relu"),
                layers.Dense(64, activation="relu"),
                keras.layers.Dense(1),
            ]
        )

        # Compile the model
        model.compile(optimizer="adam", loss="mean_absolute_error")

        model.summary()

        # Train the model
        history = model.fit(
            x_train,
            y_train,
            epochs=250,
            validation_split=0.2,
        )

        plt.plot(history.history["loss"], label="loss")
        plt.plot(history.history["val_loss"], label="val_loss")
        plt.xlabel("Epoch")
        plt.ylabel("Error [MPG]")
        plt.legend()
        plt.grid(True)
        plt.show()

        model.save(trained_model_path)

        model = keras.models.load_model(trained_model_path)

    else:
        model = keras.models.load_model(trained_model_path)

    # for every scene print the real crf at vmaf 96 and the predicted crf at vmaf 96
    for version in stats_by_version:
        crf_of_vmaf96 = 0
        closest_vmaf = 0
        # search for the crf that is the closest to vmaf 96
        for stat in stats_by_version[version]:
            if abs(stat.vmaf - 96) < abs(closest_vmaf - 96):
                closest_vmaf = stat.vmaf
                crf_of_vmaf96 = stat.basename[0]

        predicted_crf_of_vmaf96 = model.predict(
            [
                [
                    stats_by_version[version][0].basename[1],  # motion
                    stats_by_version[version][0].basename[2],  # complexity
                    # stats_by_version[version][0].basename[3],  # ref
                ]
            ],
            verbose=0,
        )[0][0]

        predicted_crf_of_vmaf96 = int(predicted_crf_of_vmaf96)

        print(
            f"{version} real: {crf_of_vmaf96} predicted: {predicted_crf_of_vmaf96} "
            f"error: {abs(predicted_crf_of_vmaf96 - crf_of_vmaf96)}%"
        )

    quit()

    # old code

    plot_dir = os.path.join(env, "plots")
    if not os.path.exists(plot_dir):
        os.mkdir(plot_dir)

    plots = []
    print(
        "real ideal crf -> resulting vmaf; predicted ideal crf -> resulting vmaf from real data; error %"
    )
    # for each scene plot "crf vs vmaf" with real data vs predicted data, unscaling the predicted data
    for chunk in analyzed_chunks:
        x_crf_of_current_chunk = [
            s.basename[0]
            for s in stats
            if s.version == chunk.path or s.version == chunk.chunk_index
        ]
        x_motion_of_current_chunk = [
            s.basename[1]
            for s in stats
            if s.version == chunk.path or s.version == chunk.chunk_index
        ]
        x_complexity_of_current_chunk = [
            s.basename[2]
            for s in stats
            if s.version == chunk.path or s.version == chunk.chunk_index
        ]
        x_reference_of_current_chunk = [
            s.basename[3]
            for s in stats
            if s.version == chunk.path or s.version == chunk.chunk_index
        ]

        y_of_current_chunk = [
            s.vmaf
            for s in stats
            if s.version == chunk.path or s.version == chunk.chunk_index
        ]

        predicted_scaled_vmaf = []
        for crf, motion, complexity, refrence in zip(
            x_crf_of_current_chunk,
            x_motion_of_current_chunk,
            x_complexity_of_current_chunk,
            x_reference_of_current_chunk,
        ):
            predicted_vmaf = (
                model.predict(
                    [
                        [
                            # crf,
                            motion,
                            complexity,
                            refrence,
                        ]
                    ],
                    verbose=0,
                )[0][0]
                * 100
            )
            predicted_scaled_vmaf.append(predicted_vmaf)

        # Unscale predicted VMAF
        predicted_vmaf = predicted_scaled_vmaf

        def test_against_val(target_value):
            closest_crf_predicted = -1
            closest_vmaf_perdicted = float("inf")
            closest_crf_real = -1
            closest_vmaf_real = float("inf")
            for crf, predicted, real in zip(
                x_crf_of_current_chunk, predicted_vmaf, y_of_current_chunk
            ):
                if abs(predicted - target_value) < abs(
                    closest_vmaf_perdicted - target_value
                ):
                    closest_vmaf_perdicted = predicted
                    closest_crf_predicted = crf
                if abs(real - target_value) < abs(closest_vmaf_real - target_value):
                    closest_vmaf_real = real
                    closest_crf_real = crf

            # within real values find the vmaf value of the closest predicted crf
            pedicted_crf_real_vmaf_value = -1
            for crf, real in zip(x_crf_of_current_chunk, y_of_current_chunk):
                if crf == closest_crf_predicted:
                    pedicted_crf_real_vmaf_value = real
                    break

            # print(
            #     f"{os.path.basename(chunk.path)} Closest crf to vmaf {target_value} is {closest_crf_predicted} (predicted) "
            #     f"and {closest_crf_real} (real), the real vmaf at {closest_crf_predicted} is {pedicted_crf_real_vmaf_value}"
            # )

            is_within_1_of_target = (
                abs(pedicted_crf_real_vmaf_value - target_value) <= 0.5
            )

            print(
                f"{os.path.basename(chunk.path)} {chunk.chunk_index}real: Crf {closest_crf_real} -> {closest_vmaf_real} VMAF,"
                f" predicted: Crf {closest_crf_predicted} -> {pedicted_crf_real_vmaf_value} VMAF;"
                f" within target {'✅' if is_within_1_of_target else '❌'}"
            )

        test_against_val(96)
        test_against_val(94)

        # Plot real vs predicted VMAF
        plt.plot(x_crf_of_current_chunk, y_of_current_chunk, label="Real VMAF")
        plt.plot(x_crf_of_current_chunk, predicted_vmaf, label="Predicted VMAF")
        plt.xlabel("CRF")
        plt.ylabel("VMAF")
        # plt.title(f"{os.path.basename(chunk.path)}")
        plt.title(f"{chunk.chunk_index}")
        plt.legend()
        # plt.show()
        # save the plot to plot with random name
        plot_path = os.path.join(plot_dir, f"{random.randint(0, 1000000)}.png")
        plt.savefig(plot_path)
        plots.append(plot_path)

        plt.clf()

    # create mosaic of plots
    mosaic = os.path.join(env, "mosaic.png")
    os.system(f"montage {' '.join(plots)} -geometry +0+0 {mosaic}")
    print(f"Mosaic of plots saved to {mosaic}")
    for plot in plots:
        os.remove(plot)



from abc import ABC


class MetricResult(ABC):
    """
    Abstract class for metric results
    """

    fps = -1
    percentile_50 = -1
    percentile_25 = -1
    percentile_10 = -1
    percentile_5 = -1
    percentile_1 = -1
    max = -1
    min = -1
    mean = -1
    harmonic_mean = -1
    std_dev = -1



"""
Testing enable-tf=0 In SvtAv1
"""
import os
from copy import deepcopy

from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.experiments.util.ExperimentUtil import (
    run_tests_across_range,
    read_report,
)

experiment_name = "Testing ALT-REF (temporally filtered) frames --enable-tf, speed 4"

test_env = os.getcwd() + "/data/temporal_filtering/"
if not os.path.exists(test_env):
    os.makedirs(test_env)

report_path = test_env + experiment_name + " CRF.json"


def run_test():
    control = EncoderSvt()
    control.speed = 4
    control.threads = 12
    control.svt_tune = 1
    control.svt_tf = 1

    enc_test = deepcopy(control)
    enc_test.svt_tf = 0

    run_tests_across_range(
        [control, enc_test],
        title=experiment_name,
        test_env=test_env,
        skip_vbr=True,
    )


def analyse():
    df = read_report(report_path)
    print(df)

    vmaf1percenttile_change_arr = []
    size_change_arr = []
    vmaf_change_arr = []
    ssim_change_arr = []

    for key, group in df.groupby("basename"):
        print(key)
        control = 0
        enc1 = 0
        for key_j, group_j in group.groupby("test_group"):
            print("")
            # avg 1%tile vmaf
            score = group_j["vmaf_percentile_1"].mean()
            size_score = group_j["size"].mean()
            vmaf_score = group_j["vmaf"].mean()
            ssim_score = group_j["ssim"].mean()
            if key_j == "control":
                control = score
                size_control = size_score
                vmaf_control = vmaf_score
                ssim_control = ssim_score
            else:
                enc1 = score
                size_enc1 = size_score
                vmaf_enc1 = vmaf_score
                ssim_enc1 = ssim_score

        # show pct of difference between control and enc1

        vmaf1percenttile_change_arr.append((enc1 - control) / control * 100)
        size_change_arr.append((size_enc1 - size_control) / size_control * 100)
        vmaf_change_arr.append((vmaf_enc1 - vmaf_control) / vmaf_control * 100)
        ssim_change_arr.append((ssim_enc1 - ssim_control) / ssim_control * 100)

    print(f"positive means enable-tf=0 is better")
    print(
        f"overall VMAF 1%tile avg: {sum(vmaf1percenttile_change_arr) / len(vmaf1percenttile_change_arr)}%"
    )
    print(f"overall size avg: {sum(size_change_arr) / len(size_change_arr)}%")
    print(f"overall VMAF avg: {sum(vmaf_change_arr) / len(vmaf_change_arr)}%")
    print(f"overall SSIM avg: {sum(ssim_change_arr) / len(ssim_change_arr)}%")

    quit()

    # split into separate df based on basename
    for key, group in df.groupby("basename"):
        print(key)
        fig, ax = plt.subplots()
        for key_j, group_j in group.groupby("test_group"):
            group_j.plot(
                ax=ax,
                kind="scatter",
                x="size",
                y="vmaf_percentile_1",
                label=key_j,
                color=("red" if key_j == "control" else "blue"),
            )

        # 3d plot rate vmaf_percentile_1 size colored by test_group
        # fig = plt.figure()
        # ax = fig.add_subplot(111, projection="3d")
        # for key_j, group_j in group.groupby("test_group"):
        #     ax.scatter(
        #         group_j["rate"],
        #         group_j["vmaf_percentile_1"],
        #         group_j["size"],
        #         label=key_j,
        #         color=("red" if key_j == "control" else "blue"),
        #     )
        # ax.set_xlabel("rate")
        # ax.set_ylabel("vmaf_percentile_1")
        # ax.set_zlabel("size")

        plt.title(key)
        plt.savefig(test_env + key + ".png")


if __name__ == "__main__":
    run_test()
    pass



from abc import ABC, abstractmethod

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.encoder import Encoder
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.scene.chunk import ChunkObject


class FinalEncodeStep(ABC):
    """
    Preforms the final (usually longer) encoding
    """

    @abstractmethod
    def run(
        self,
        enc: Encoder,
        chunk: ChunkObject,
        ctx: AlabamaContext,
        encoded_a_frame,
    ) -> EncodeStats:
        pass

    @abstractmethod
    def dry_run(self, enc: Encoder, chunk: ChunkObject) -> str:
        pass



import os
import shutil
from multiprocessing import Pool

from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.encoder.impl.X264 import EncoderX264
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.scene.sequence import ChunkSequence


def run_tune(a):
    tune, enc, path, vmaf_options = a
    enc.x264_tune = tune
    enc.output_path = os.path.join(path, f"{tune}{enc.get_chunk_file_extension()}")
    stats = enc.run(metric_to_calculate=Metric.VMAF, metric_params=vmaf_options)
    print(f"{tune}: {stats.bitrate}kb/s {stats.metric_results.harmonic_mean}vmaf")
    return tune, stats.metric_results.harmonic_mean, stats.bitrate


def get_ideal_x264_tune(ctx: AlabamaContext, sequence: ChunkSequence):
    """
    Picks the ideal x264 tune for the sequence, based on the vmaf per bitrate
    Very ad-hoc, but it ~works
    :param ctx:
    :param sequence:
    :return:
    """

    if isinstance(ctx.prototype_encoder, EncoderX264):
        ctx.prototype_encoder.x264_tune = ctx.get_kv().get("x264_tune", "value")

        if ctx.prototype_encoder.x264_tune is None:
            print("picking x264 tune")
            tunes = ["animation", "film", "grain"]
            enc = ctx.get_encoder()
            enc.crf = 23
            enc.speed = 4

            path = os.path.join(ctx.temp_folder, "adapt", "x264_tune/")

            if not os.path.exists(path):
                os.makedirs(path)

            best = []

            for test_chunk in sequence.get_test_chunks_out_of_a_sequence(
                random_pick_count=2
            ):
                enc.chunk = test_chunk

                # get (tune, stats) tuples in a parallel fashion
                with Pool() as p:
                    runs = p.map(
                        run_tune,
                        [
                            (tune, enc.clone(), path, ctx.get_vmaf_options())
                            for tune in tunes
                        ],
                    )
                    # and close the pool
                    p.close()
                    p.join()

                # pick the tune with the biggest vmaf per bitrate with a little bias towards vmaf
                # runs_calculated = [
                #     (tune, ((vmaf / bitrate) + (vmaf / 100)))
                #     for tune, vmaf, bitrate in runs
                # ]
                # runs_calculated.sort(key=lambda x: x[1])
                # best_tune = runs_calculated[-1][0]

                # pick the tune with the biggest vmaf
                runs.sort(key=lambda x: x[1])
                best_tune = runs[-1][0]

                best.append(best_tune)

            # pick the most common tune from the best tunes
            ctx.prototype_encoder.x264_tune = max(set(best), key=best.count)
            print(f"picked {ctx.prototype_encoder.x264_tune} as x264 tune")
            ctx.get_kv().set("x264_tune", "value", ctx.prototype_encoder.x264_tune)
            shutil.rmtree(path)
        else:
            print(f"setting x264 tune to {ctx.prototype_encoder.x264_tune} from cache")

    return ctx



"""
Testing svtav1 speed 2v3v4 and which one is the best tradeoff
"""
import os

from alabamaEncode.encoder.impl.Svtenc import EncoderSvt
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.experiments.util.ExperimentUtil import get_test_files
from alabamaEncode.scene.chunk import ChunkObject

if __name__ == "__main__":
    paths = get_test_files()[:6]

    for path in paths:
        print(f"\n\nDoing: {path}")
        chunk = ChunkObject(path=path, first_frame_index=0, last_frame_index=200)

        crope_stringe = ""

        svtenc = EncoderSvt()

        svtenc.svt_chroma_thing = 0
        svtenc.keyint = 999
        svtenc.svt_open_gop = False

        test_env = "./tstVBR" + path.split("/")[-1].split(".")[0] + "/"
        # shutil.rmtree(test_env, ignore_errors=True)
        if not os.path.exists(test_env):
            os.mkdir(test_env)

        svtenc.rate_distribution = EncoderRateDistribution.VBR
        svtenc.bitrate = 1500
        svtenc.passes = 3
        svtenc.chunk = chunk
        svtenc.current_scene_index = 0
        svtenc.threads = 12
        svtenc.video_filters = crope_stringe
        svtenc.grain_synth = 3

        if "liveAction_normal" in path:
            svtenc.bitrate = 1000
        elif "liveAction_highMotion" in path:
            svtenc.bitrate = 2000
        elif "Animation" in path:
            svtenc.bitrate = 1500
        elif "stopmotion" in path:
            svtenc.bitrate = 3000
        elif "liveAction_4k" in path:
            svtenc.bitrate = 4000
        elif "liveaction_bright" in path:
            svtenc.bitrate = 1000

        svtenc.rate_distribution = EncoderRateDistribution.VBR
        svtenc.passes = 3

        print(
            f"| _Vbr {svtenc.bitrate}_ |  time taken  | kpbs | vmaf  | BD Change % | time Change % |\n"
            "|----------|:------:|:----:|:-----:|:-----------:|:---:|"
        )
        four_dbrate = -1
        four_time = -1
        for speed in [4, 3, 2]:
            print(f"\nRunning speed {speed}")
            svtenc.output_path = f"{test_env}speed{speed}.ivf"
            svtenc.speed = speed
            print(svtenc.get_encode_commands())
            quit()
            stats = alabamaEncode.core.run(
                override_if_exists=False, metric_to_calculate=Metrics.VMAF
            )

            curr_db_rate = stats.size_kB / stats.vmaf
            if speed == 4:
                four_time = stats.time_encoding
                four_dbrate = curr_db_rate

            change_from_zero = (curr_db_rate - four_dbrate) / four_dbrate * 100
            change_from_zero_time = (stats.time_encoding - four_time) / four_time * 100
            print(
                f"| {speed} |"
                f" {stats.time_encoding}s |"
                f" {stats.bitrate} "
                f"| {stats.vmaf} | "
                f" {change_from_zero}% |"
                f" {change_from_zero_time}% |"
            )



import json
import os
import random

from torf import Torrent
from tqdm import tqdm

from alabamaEncode.cli.cli_setup.autopaths import is_title_movie, parse_movie_title
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.path import PathAlabama


def print_stats(
    output_folder: str,
    output: str,
    title: str,
):
    result_file = f"{output_folder}/stat.txt"
    if os.path.exists(result_file):
        os.remove(result_file)

    def ps(s):
        print(s)
        with open(result_file, "a") as stat_file:
            stat_file.write(s + "\n")

    output_file = PathAlabama(output)

    def sizeof_fmt(num, suffix="B"):
        for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
            if abs(num) < 1024.0:
                return f"{num:3.1f}{unit}{suffix}"
            num /= 1024.0
        return f"{num:.1f}Yi{suffix}"

    # Title: <Title> <episode & seson if applicable>
    # IMDB: <id>
    # Runtime: hh:mm:ss
    # Size: 0.00 GB|MB
    #
    # Video (<codec>):
    # <width>x<height> - <8|10|12>-bit <HDR> - 0000 kbps
    #
    # Audio (<codec>):
    # English <channel layout eg 7.1> - 000 kbps
    #
    # Subtitles (<number of subs tracks>x <sub title codec eg SRT>):
    # <All subtitle langs comma separated>
    #
    # Encoder / Settings: <encoder used eg svt-av1-psy>, Preset <speed used eg 4>, <rd mode eg target-quality VMAF 94>, < gs used if applicable eg Grain Synth 5>
    #
    # <if av1 is used this message:
    #     Compatibility: Windows/Android: VLC or MPV - Oculus: SkyboxVR 1.1.0 or later - TVs: Most smart TVs from 2020 or newer>

    size_bytes = os.path.getsize(output_file.get()) * 8

    width, height = Ffmpeg.get_width(output_file), Ffmpeg.get_height(output_file)

    bitdepth = Ffmpeg.get_bit_depth(output_file)

    hdr = Ffmpeg.is_hdr(output_file)

    vid_bps, audio_bps = Ffmpeg.get_source_bitrates(output_file)

    tracks = Ffmpeg.get_tracks(output_file)

    # running under one vid track assumption
    video_tracks = []
    # running under one audio track assumption
    audio_tracks = []
    sub_tracks = []

    for track in tracks:
        if track["codec_type"] == "subtitle":
            sub_tracks.append(track)
        elif track["codec_type"] == "audio":
            audio_tracks.append(track)
        elif track["codec_type"] == "video":
            video_tracks.append(track)
    lenght = Ffmpeg.get_video_length(output_file, sexagesimal=True)
    vid_codec = video_tracks[0]["codec_name"].upper()

    if len(audio_tracks) == 0:
        audio_codec = "N/A"
    else:
        audio_codec = audio_tracks[0]["codec_name"].upper()
    if len(sub_tracks) == 0:
        sub_codec = "N/A"
    else:
        sub_codec = sub_tracks[0]["codec_name"].upper()

    sub_langs_list = []
    for sub_track in sub_tracks:
        lang = "N/A"
        if "title" in sub_track["tags"]:
            lang = sub_track["tags"]["title"]
        elif "language" in sub_track["tags"]:
            lang = sub_track["tags"]["language"]
        sub_langs_list.append(lang)

    print("\n")

    if is_title_movie(title):
        title, year = parse_movie_title(title)
        ps(f"Title: {title}")
    else:
        ps(f"Title: {title}")

    ps("IMDB: <filled manually>")

    ps(f"Runtime: {lenght}")

    ps(f"Size: {sizeof_fmt(size_bytes)}")
    ps("\n")

    ps(f"Video ({vid_codec}):")
    ps(
        f"{width}x{height} - {bitdepth}-bit {'HDR' if hdr else 'SDR'} - {vid_bps // 1000} kbps"
    )
    ps("\n")

    ps(f"Audio ({audio_codec}):")
    for audio_track in audio_tracks:
        lang = "N/A"
        if "language" in audio_track["tags"]:
            lang = audio_track["tags"]["language"]
        ps(f"{lang} {audio_track['channel_layout']} - {audio_bps // 1000} kbps")

    ps("\n")

    ps(f"Subtitles ({len(sub_tracks)}x {sub_codec}):")

    ps(", ".join(sub_langs_list))

    ps("\n")

    ps("Encoder / Settings: <FILL MANUALLY>")

    if vid_codec == "AV1":
        ps(
            "Compatibility: Windows/Android: VLC or MPV - Oculus: SkyboxVR 1.1.0 or later - TVs: Most smart TVs from 2020 or newer"
        )

    ps("\n")


def generate_previews(input_file: str, output_folder: str):
    total_length = Ffmpeg.get_video_length(PathAlabama(input_file))  # lenght in seconds

    # one preview every 5 minutes, min 1 max 4
    num_previews = max(1, min(4, int(total_length / 300)))

    offsets = []

    for i in range(num_previews):
        offsets.append(int(random.uniform(0, total_length)))

    is_av1 = Ffmpeg.get_codec(PathAlabama(input_file)) == "av1"

    for i, offset in tqdm(
        enumerate(offsets), desc="Generating previews", total=num_previews
    ):
        # create additional avif stream copy previews
        if is_av1:
            run_cli(
                f'{get_binary("ffmpeg")} -y -ss {offset} -i "{input_file}" -t 5 '
                f'-c copy "{output_folder}/preview_{i}.avif"'
            )
        png_preview_path = f"{output_folder}/preview_{i}.png"
        jpeg_preview_path = f"{output_folder}/preview_{i}.jpeg"
        run_cli(
            f'{get_binary("ffmpeg")} -y -ss {offset}'
            f' -i "{input_file}" -v:frames 1 -pix_fmt rgb24 "{png_preview_path}"'
        )
        run_cli(
            f'cat "{png_preview_path}" | {get_binary("cjpeg")} -q 95 -tune-psnr -optimize -progressive > '
            f'"{jpeg_preview_path}"'
        )
        run_cli(f'{get_binary("ect")} "{png_preview_path}"')
        run_cli(f'{get_binary("ect")} "{jpeg_preview_path}"')
        # try:
        #     pass
        # except BinaryNotFound:
        #     print("cjpeg or ect not found, skipping jpeg preview/png optimization")


if __name__ == "__main__":
    # register_bin("ect", "/home/kokoniara/.local/opt/ect")
    # generate_previews(
    #     "/home/kokoniara/dev/VideoSplit/test_codes/lessons_in_meth_fast13_slow4/out.mp4",
    #     "/home/kokoniara/dev/VideoSplit/test_codes/lessons_in_meth_fast13_slow4/",
    # )
    tracks = Ffmpeg.get_tracks(
        PathAlabama(
            "/home/kokoniara/showsEncode/Foundation (2021)/s2/Foundation.2021.S02.1080p.AV1.OPUS.SouAV1R/Foundation.2021.S02E08.1080p.AV1.OPUS.SouAV1R.webm"
        )
    )
    # pretty print the dictionary
    print(json.dumps(tracks, sort_keys=True, indent=4))


def create_torrent_file(video: str, encoder_name: str, output_folder: str):
    trackers = [
        "udp://tracker.opentrackr.org:1337/announce",
        "https://tracker2.ctix.cn:443/announce",
        "https://tracker1.520.jp:443/announce",
        "udp://opentracker.i2p.rocks:6969/announce",
        "udp://tracker.openbittorrent.com:6969/announce",
        "http://tracker.openbittorrent.com:80/announce",
        "udp://open.demonii.com:1337/announce",
        "udp://open.stealth.si:80/announce",
        "udp://exodus.desync.com:6969/announce",
        "udp://tracker.torrent.eu.org:451/announce",
        "udp://tracker.moeking.me:6969/announce",
        "udp://explodie.org:6969/announce",
        "udp://tracker.opentrackr.org:1337/announce",
        "http://tracker.openbittorrent.com:80/announce",
        "udp://opentracker.i2p.rocks:6969/announce",
        "udp://tracker.internetwarriors.net:1337/announce",
        "udp://tracker.leechers-paradise.org:6969/announce",
        "udp://coppersurfer.tk:6969/announce",
        "udp://tracker.zer0day.to:1337/announce",
    ]

    print("Creating torrent file")

    t = Torrent(path=video, trackers=trackers)
    t.comment = f"Encoded by {encoder_name}"

    t.private = False

    t.generate()

    torrent_file_path = os.path.join(output_folder, "torrent.torrent")

    if os.path.exists(torrent_file_path):
        os.remove(torrent_file_path)
    t.write(torrent_file_path)



from alabamaEncode.core.context import AlabamaContext
from alabamaEncode.scene.sequence import ChunkSequence


def setup_denoise(ctx: AlabamaContext, sequence: ChunkSequence):
    """
    Sets up a denoise filter
    """
    if ctx.simple_denoise:
        print("Using simple denoise")

        # ctx.prototype_encoder.film_grain_denoise = 1
        vf = ctx.prototype_encoder.video_filters.split(",")
        vf = [_x for _x in vf if _x != ""]
        vf.append("atadenoise")
        ctx.prototype_encoder.video_filters = ",".join(vf)

    return ctx



import os

from alabamaEncode.conent_analysis.opinionated_vmaf import get_vmaf_list
from alabamaEncode.conent_analysis.refine_step import RefineStep
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.scene.concat import VideoConcatenator


class MutliResPackage(RefineStep):
    def __call__(self, ctx, sequence):
        final_paths = ctx.get_kv().get("multires_final_paths", "final_paths")
        if final_paths is None:
            print("Not packaging, final paths not done")
            return

        codec = ctx.prototype_encoder.get_codec()
        vmafs = get_vmaf_list(codec)

        for chunk in sequence.chunks:
            for vmaf in vmafs:
                chunk_vmaf_path = ctx.get_kv().get(
                    "multires_final_paths", f"{chunk.chunk_index}_{vmaf}"
                )
                if chunk_vmaf_path is None or not os.path.exists(chunk_vmaf_path):
                    print("Not packaging, final paths not done")
                    return

        concated_files = []

        ctc = VideoConcatenator(
            output="",
            file_with_audio=ctx.input_file,
            start_offset=ctx.start_offset,
            end_offset=ctx.end_offset,
            title=ctx.get_title(),
            encoder_name=ctx.encoder_name,
            mux_audio=False,
            subs_file=[],
        )

        for path in final_paths:
            vmaf_target = path["vmaf_target"]
            print(f"\n vmaf_target: {vmaf_target}")
            # concat
            out_path = os.path.join(
                ctx.output_folder,
                f"vmaf{vmaf_target}_{path['res']}.mp4",
            )
            ctc.output = out_path
            concated_files += [out_path]

            ctc.files = []
            for c in sequence.chunks:
                chunk_vmaf_path = ctx.get_kv().get(
                    "multires_final_paths", f"{c.chunk_index}_{vmaf_target}"
                )
                ctc.files += [chunk_vmaf_path]

            ctc.concat_videos()

        print("\n encoding audio tracks")
        audio_track_96_path = os.path.join(
            ctx.output_folder,
            f"audio_track_96.mp4",
        )
        ctc.audio_param_override = "-c:a libopus -ac 2 -b:a 96k -vbr on"
        ctc.output = audio_track_96_path
        ctc.audio_only = True
        ctc.concat_videos()

        audio_track_128_path = os.path.join(
            ctx.output_folder,
            f"audio_track_128.mp4",
        )
        ctc.audio_param_override = "-c:a libopus -ac 2 -b:a 128k -vbr on"
        ctc.output = audio_track_128_path
        ctc.audio_only = True
        ctc.concat_videos()

        # example cli:
        # packager \
        # in=64k.webm,stream=audio,output=64.webm \
        # in=256k.webm,stream=audio,output=256.webm \
        # in=og_480p.webm,stream=video,output=480p.webm \
        # in=og_1080p.webm,stream=video,output=1080p.webm \
        # in=og_480p.webm,stream=video,output=480p_trick.webm,trick_play_factor=1 \
        # in=og_1080p.webm,stream=video,output=1080p_trick.webm,trick_play_factor=1 \
        # --mpd_output ./stream/stream.mpd

        files_cli = ""
        for file in concated_files:
            files_cli += f" in={file},stream=video,output={file} "

        trick_path = os.path.join(
            ctx.output_folder,
            f"trick.mp4",
        )
        trick = f" in={concated_files[-1]},stream=video,output={trick_path},trick_play_factor=1 "
        audio = (
            f" in={audio_track_96_path},stream=audio,output={audio_track_96_path} "
            f" in={audio_track_128_path},stream=audio,output={audio_track_128_path}"
        )

        print("\n\n")
        final_cli = (
            f"{get_binary('packager')} {files_cli} {trick} {audio}"
            f" --segment_duration 3 "
            f" --mpd_output {ctx.output_file}"
        )
        print("Running final cli:")
        print(final_cli)
        os.system(final_cli)

        stats_content = ""
        for path in final_paths:
            out_path = PathAlabama(
                os.path.join(
                    ctx.output_folder,
                    f"vmaf{path['vmaf_target']}_{path['res']}.mp4",
                )
            )
            bitrate = int(Ffmpeg.get_total_bitrate(out_path) / 1000)
            stats_content += (
                f"vmaf {path['vmaf_target']} res {path['res']} bitrate {bitrate}\n"
            )

        # audio track
        bitrate = int(Ffmpeg.get_total_bitrate(PathAlabama(audio_track_96_path)) / 1000)
        stats_content += f"audio bitrate {bitrate}k\n"

        bitrate = int(
            Ffmpeg.get_total_bitrate(PathAlabama(audio_track_128_path)) / 1000
        )
        stats_content += f"audio bitrate {bitrate}k\n"

        with open(os.path.join(ctx.output_folder, "stats.txt"), "w") as f:
            f.write(stats_content)



from alabamaEncode.core.util.bin_utils import get_binary, register_bin
from alabamaEncode.core.util.cli_executor import run_cli_parallel
from alabamaEncode.core.util.path import PathAlabama


from alabamaEncode.metrics.exception import Ssimu2Exception
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.metrics.options import MetricOptions
from alabamaEncode.metrics.result import MetricResult
from alabamaEncode.scene.chunk import ChunkObject


class Ssimu2Options(MetricOptions):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


def calc_ssimu2(
    chunk: ChunkObject,
    ssimu2_options: Ssimu2Options,
):
    assert ssimu2_options is not None

    from alabamaEncode.metrics.calculate import get_input_pipes

    owo = get_input_pipes(chunk=chunk, options=ssimu2_options)

    ref_pipe = owo["ref_pipe"]
    dist_pipe = owo["dist_pipe"]
    ref_command = owo["ref_command"]
    dist_command = owo["dist_command"]

    main_command = f"{get_binary('ssimulacra2_rs')} video --frame-threads 4 {ref_pipe} {dist_pipe} "

    cli_results = run_cli_parallel(
        [
            ref_command,
            dist_command,
            main_command,
        ]
    )

    from alabamaEncode.metrics.calculate import cleanup_input_pipes

    cleanup_input_pipes(owo)

    for c in cli_results:
        try:
            c.verify()
        except RuntimeError as e:
            raise Ssimu2Exception(
                f"Could not run ssimu2 command: {e}, {[c.output for c in cli_results]}"
            )

    return Ssimu2Result(cli_results[2].output)


class Ssimu2Result(MetricResult):
    def __init__(self, cli_out: str):
        for line in cli_out.split("\n"):
            if "Mean" in line:
                self.mean = float(line.split(":")[1])
            if "Median" in line:
                self.percentile_50 = float(line.split(":")[1])
            if "Std Dev" in line:
                self.std_dev = float(line.split(":")[1])
            if "5th Percentile" in line:
                self.percentile_5 = float(line.split(":")[1])
            if "95th Percentile" in line:
                self.percentile_95 = float(line.split(":")[1])

    def __str__(self):
        return f"{self.mean}"

    def __repr__(self):
        return (
            f"Ssimu2Result(mean={self.mean},"
            f" prct_95={self.percentile_95},"
            f" prct_5={self.percentile_5},"
            f" std_dev={self.std_dev})"
        )


# test
if __name__ == "__main__":
    print("calcing ssimu2")
    register_bin(
        "ssimulacra2_rs",
        "/home/kokoniara/.local/opt/ssimulacra2_rs",
    )
    from alabamaEncode.metrics.calculate import calculate_metric

    result: Ssimu2Result = calculate_metric(
        reference_path=PathAlabama("/mnt/data/objective-1-fast/MINECRAFT_60f_420.y4m"),
        distorted_path=PathAlabama(
            "/mnt/data/objective-1-fast/minecruft_test_encode_h264.mkv"
        ),
        options=Ssimu2Options(threads=10),
        metric=Metric.SSIMULACRA2,
    )
    print(result.mean)
    print(result.percentile_1)
    print(result.__repr__())



import copy
import os
import random
from typing import List

from tqdm.asyncio import tqdm

from alabamaEncode.core.util.kv import AlabamaKv
from alabamaEncode.scene.chunk import ChunkObject


class ChunkSequence:
    """
    A sequence of chunks.
    """

    def __init__(self, chunks: List[ChunkObject]):
        self.chunks = chunks
        self.input_file = ""

    def get_specific_chunk(self, index: int) -> ChunkObject:
        return self.chunks[index]

    def __len__(self):
        return len(self.chunks)

    def __getitem__(self, index):
        return self.chunks[index]

    def dump_json(self) -> str:
        """
        Dumps the sequence to json
        :return: string
        """
        import json

        d = {"chunks": [c.__dict__ for c in self.chunks], "input_file": self.input_file}
        return json.dumps(d)

    def load_json(self, json_load: str) -> "ChunkSequence":
        """
        Loads the sequence from json
        :param json_load: path to the json file to load
        :return: self
        """
        import json

        d = json.loads(json_load)

        self.chunks = [ChunkObject() for _ in d["chunks"]]
        for i, c in enumerate(d["chunks"]):
            self.chunks[i].__dict__ = c

        self.input_file = d["input_file"]
        return self

    def setup_paths(self, temp_folder: str, extension: str):
        """
        sets up the paths for the chunks, in the appropriate temp folder, and with the appropriate extension
        :param extension: .ivf .mkv etc.
        :param temp_folder: the temp folder to put the chunks in
        :return: Void
        """
        os.makedirs(os.path.join(temp_folder, "chunks"), exist_ok=True)
        for c in self.chunks:
            # /home/user/encode/show/temp/1.ivf
            # or
            # /home/user/encode/show/temp/1.mkv
            c.chunk_path = os.path.join(
                temp_folder, f"chunks/{c.chunk_index}{extension}"
            )

    def get_test_chunks_out_of_a_sequence(
        self, random_pick_count: int = 7
    ) -> List[ChunkObject]:
        """
        Get an equally distributed list of chunks from a sequence for testing, does not modify the original sequence
        :param random_pick_count: Number of random chunks to pick
        :return: List of Chunk objects
        """
        chunks_copy: List[ChunkObject] = copy.deepcopy(self.chunks)
        chunks_copy = chunks_copy[
            int(len(chunks_copy) * 0.2) : int(len(chunks_copy) * 0.8)
        ]

        if len(chunks_copy) > 10:
            # bases on length, remove every x scene from the list so its shorter
            chunks_copy = chunks_copy[:: int(len(chunks_copy) / 10)]

        random.shuffle(chunks_copy)
        chunks = chunks_copy[:random_pick_count]

        if len(chunks) == 0:
            print("Failed to shuffle chunks for analysis, using all")
            chunks = self.chunks

        return copy.deepcopy(chunks)

    def sequence_integrity_check(self, kv: AlabamaKv = None) -> bool:
        """
        checks the integrity of the chunks, and removes any that are invalid, and see if all are done
        :param kv: AlabamaKv object, for checking if a chunk is valid from cache
        :return: true if there are broken chunks / not all chunks are done
        """

        seq_chunks = list(self.chunks)
        total_chunks = len(self.chunks)
        invalid_chunks: List[ChunkObject or None] = []

        for chunk in tqdm(seq_chunks, desc="Checking files", unit="file"):
            if not chunk.is_done(kv=kv, length_of_sequence=total_chunks):
                invalid_chunks.append(chunk)

        del_count = 0

        invalid_chunks: List[ChunkObject] = [
            chunk for chunk in invalid_chunks if chunk is not None
        ]

        if len(invalid_chunks) > 0:
            for c in invalid_chunks:
                if os.path.exists(c.chunk_path):
                    os.remove(c.chunk_path)
                    print(f"Deleted invalid file {c.chunk_path}")
                    del_count += 1
            return True

        if del_count > 0:
            print(f"Deleted {del_count} invalid files 😂")

        undone_chunks_count = 0

        for c in self.chunks:
            if not c.chunk_done:
                undone_chunks_count += 1

        if undone_chunks_count > 0:
            print(
                f"Only {total_chunks - undone_chunks_count}/{total_chunks} chunks are done 😏"
            )
            return True

        print("All chunks passed integrity checks 🤓")
        return False



from alabamaEncode.core.context import AlabamaContext

if __name__ == "__main__":
    ctx = AlabamaContext()

    ctx.use_celery = "KAWAI"

    json = ctx.to_json()

    print(json)

    from_json = AlabamaContext().from_json(json)

    print(from_json)



from alabamaEncode.ai_vmaf.aom_firstpass import aom_extract_firstpass_data
from alabamaEncode.scene.chunk import ChunkObject

if __name__ == "__main__":
    chunk = ChunkObject(
        path="/mnt/data/objective-1-fast/Netflix_RollerCoaster_1280x720_60fps_8bit_420_60f.y4m"
    )

    aom_extract_firstpass_data(chunk)



from alabamaEncode.conent_analysis.opinionated_vmaf import (
    get_vmaf_list,
    convexhull_get_crf_range,
    convexhull_get_resolutions,
)
from alabamaEncode.conent_analysis.refine_step import RefineStep
from alabamaEncode.encoder.codec import Codec


class MutliResTrellis(RefineStep):
    def __call__(self, ctx, sequence):
        if ctx.get_kv().get("multires_final_paths", "final_paths") is not None:
            print("Not calculating trellis, already done")
            return

        codec = ctx.prototype_encoder.get_codec()
        vmafs = get_vmaf_list(codec)
        crf_low, crf_high = convexhull_get_crf_range(codec)
        crf_range = list(range(crf_low, crf_high, 2))

        resolutions = convexhull_get_resolutions(codec)
        resolutions = [r.split(":")[0] for r in resolutions]

        # vmafs_values = []
        # for res in resolutions:
        #     for vmaf_target in vmafs:
        #         for chunk in sequence.chunks:
        #             for crf in crf_range:
        #                 data = ctx.get_kv().get(
        #                     "multi_res_candidates",
        #                     f"{chunk.chunk_index}_{res}_{crf}",
        #                 )
        #                 if data is not None:
        #                     vmafs_values.append(data["vmaf"])

        # scaled_max = max(vmafs_values)
        # pick based on percentiles
        # scaled_max = sorted(vmafs_values)[int(len(vmafs_values) * 0.85)]

        # max_target_vmaf = max(vmafs)
        # if scaled_max < max_target_vmaf:
        #     data_min = min(vmafs)
        #     range_unscaled = max_target_vmaf - min(vmafs)
        #     range_scales = scaled_max - data_min
        #     vmafs_scales = [(data - data_min) / range_unscaled for data in vmafs]
        #     print(vmafs_scales)
        #     vmafs = [data_min + (range_scales * data) for data in vmafs_scales]
        #
        #     print(
        #         f"Compressing vmaf range to: {data_min} - {scaled_max}; final vmaf targets: {vmafs}"
        #     )

        fps = sequence.chunks[0].framerate

        trellis_paths = {}
        for res in resolutions:
            for vmaf_target in vmafs:
                current_path = []
                print(f"Calculating trellis for res: {res} vmaf_target: {vmaf_target}")
                for chunk in sequence.chunks:
                    all_crfs = []
                    for crf in crf_range:
                        data = ctx.get_kv().get(
                            "multi_res_candidates",
                            f"{chunk.chunk_index}_{res}_{crf}",
                        )
                        if data is None:
                            continue
                        data = {
                            "vmaf": data["vmaf"],
                            "crf": data["crf"],
                            "bitrate": data["bitrate"],
                            "file": data["file"],
                            "res": data["res"],
                            "index": chunk.chunk_index,
                            "length": chunk.get_lenght(),
                        }
                        all_crfs.append(data)

                    # sort by vmaf error from the target vmaf

                    best_fit = all_crfs[0]
                    if ctx.prototype_encoder.get_codec() == Codec.av1:

                        def get_s(x):
                            frames = x["length"] * fps
                            bitrate = x["bitrate"]  # in kbps

                            # bitrate component
                            bitrate_weight = (bitrate / 650) ** 0.4 - 1
                            bitrate_weight *= 2

                            # Exponential component for the number of frames
                            size_kb = (frames * bitrate) / 1000
                            overall_size_weight = (size_kb / 250) ** 0.5
                            overall_size_weight = max(overall_size_weight, 1)

                            # Combine the components
                            combined_weight = bitrate_weight * overall_size_weight

                            # Ensure the combined weight is not negative
                            combined_weight = max(combined_weight, 0)
                            return abs(vmaf_target - x["vmaf"]) + combined_weight

                        all_crfs.sort(key=lambda x: get_s(x))
                        best_fit = all_crfs[0]

                    else:
                        all_crfs.sort(key=lambda x: abs(x["vmaf"] - vmaf_target))
                        best_fit = all_crfs[0]

                    current_path.append(best_fit)

                if f"{vmaf_target}" not in trellis_paths:
                    trellis_paths[f"{vmaf_target}"] = {}
                trellis_paths[f"{vmaf_target}"][f"{res}"] = current_path

        # now the filtering part, we need to pick the len(vmafs) best paths, one for each quality target
        vmafs_reverse = vmafs[::-1]
        prev_bitrate = -1
        final_paths = []
        for vmaf_target in vmafs_reverse:
            print("\n")
            trys_in_vmaf_group = trellis_paths[f"{vmaf_target}"]
            candidates = []
            for res in resolutions:
                paths_in_res = trys_in_vmaf_group[res]
                vmaf_errors = []
                bits = 0  # kb of all chunks together
                lengths = 0  # seconds of all chunks together
                for chunk in paths_in_res:
                    vmaf_errors.append(abs(chunk["vmaf"] - vmaf_target))
                    # get bits from bitrate and length filed
                    bits += chunk["bitrate"] * chunk["length"]
                    lengths += chunk["length"]

                vmaf_error_avg = sum(vmaf_errors) / len(vmaf_errors)
                bitrate = bits / lengths
                print(
                    f"vmaf target: {vmaf_target};   res: {res};   vmaf_error_avg: {vmaf_error_avg};  bitrate: {bitrate}"
                )
                candidates.append(
                    {
                        "vmaf_error_avg": vmaf_error_avg,
                        "bitrate": bitrate,
                        "res": res,
                        "paths": paths_in_res,
                    }
                )

            if prev_bitrate == -1:
                candidates.sort(key=lambda x: x["vmaf_error_avg"])
            else:
                candidates.sort(
                    key=lambda x: x["vmaf_error_avg"]
                    + min(abs(x["bitrate"] - prev_bitrate), 0)
                )

            best_candidate = candidates[0]
            print(
                f"best candidate for vmaf target {vmaf_target} is res: {best_candidate['res']} "
                f"with vmaf_error_avg: {best_candidate['vmaf_error_avg']}"
            )
            prev_bitrate = best_candidate["bitrate"]
            best_candidate = {
                **best_candidate,
                "vmaf_target": vmaf_target,
            }
            final_paths.append(best_candidate)

        # a = {
        #     "res": "720",
        #     "vmaf": "95",
        #     "chunks": [
        #         {"chunk_index": "0", "crf": "11"},
        #         {"chunk_index": "1", "crf": "12"},
        #     ],
        # }

        for path in final_paths:
            obj = {"res": path["res"], "chunks": []}

            for chunk in path["paths"]:
                obj["chunks"].append(
                    {
                        "chunk_index": chunk["index"],
                        "crf": chunk["crf"],
                    }
                )

            ctx.get_kv().set(
                "multires_trellis",
                f"{path['vmaf_target']}",
                obj,
            )

        ctx.get_kv().set("multires_final_paths", "final_paths", final_paths)



from collections import namedtuple

# https://github.com/ticapix/python-y4m

class Frame(namedtuple('Frame', ['buffer', 'headers', 'count'])):
    def __repr__(self):
        return '<frame %d: %dx%d>' % (self.count, self.headers['H'], self.headers['W'])

class Reader(object):
    def __init__(self, callback, verbose=False):
        self._callback = callback
        self._stream_headers = None
        self._data = bytes()
        self._count = 0
        self._verbose = verbose

    def _print(self, *args):
        if self._verbose:
            print('Y4M Reader:', ' '.join([str(e) for e in args]))

    def decode(self, data):
        assert isinstance(data, bytes)
        self._data += data
        if self._stream_headers is None:
            self._decode_stream_headers()
            if self._stream_headers is not None:
                self._print('detected stream with headers:', self._stream_headers)
        if self._stream_headers is not None:
            frame = self._decode_frame()
            while frame is not None:
                self._print(frame, 'decoded')
                self._callback(frame)
                frame = self._decode_frame()

    def _frame_size(self):
        if self._stream_headers['C'].startswith('420'):
            return self._stream_headers['W'] * self._stream_headers['H'] * 3 // 2
        elif self._stream_headers['C'].startswith('422'):
            return self._stream_headers['W'] * self._stream_headers['H'] * 2
        elif self._stream_headers['C'].startswith('444'):
            return self._stream_headers['W'] * self._stream_headers['H'] * 3
        raise f"only support I420, I422, I444 fourcc (not {self._stream_headers['C']})"

    def _decode_frame(self):
        if len(self._data) < self._frame_size():  # no point trying to parse
            return None
        toks = self._data.split(b'\n', 1)
        if len(toks) == 1:  # need more data
            self._print('weird: got plenty of data but no frame header found')
            return None
        headers = toks[0].split(b' ')
        # assert headers[0] == b'FRAME', 'expected FRAME (got %r)' % headers[0]
        frame_headers = self._stream_headers.copy()
        for header in headers[1:]:
            header = header.decode('ascii')
            frame_headers[header[0]] = header[1:]
        if len(toks[1]) < self._frame_size():  # need more data
            return None
        yuv = toks[1][0:self._frame_size()]
        self._data = toks[1][self._frame_size():]
        self._count += 1
        return Frame(yuv, frame_headers, self._count - 1)

    def _decode_stream_headers(self):
        toks = self._data.split(b'\n', 1)
        if len(toks) == 1:  # buffer all header data until eof
            return
        self._stream_headers = {}
        self._data = toks[1]  # save the beginning of the stream for later
        headers = toks[0].split(b' ')
        assert headers[0] == b'YUV4MPEG2', 'unknown type %s' % headers[0]
        for header in headers[1:]:
            header = header.decode('ascii')
            self._stream_headers[header[0]] = header[1:]
        assert 'W' in self._stream_headers, 'No width header'
        assert 'H' in self._stream_headers, 'No height header'
        assert 'F' in self._stream_headers, 'No frame-rate header'
        self._stream_headers['W'] = int(self._stream_headers['W'])
        self._stream_headers['H'] = int(self._stream_headers['H'])
        self._stream_headers['F'] = [int(n) for n in self._stream_headers['F'].split(':')]
        if 'A' in self._stream_headers:
            self._stream_headers['A'] = [int(n) for n in self._stream_headers['A'].split(':')]
        if 'C' not in self._stream_headers:
            self._stream_headers['C'] = '420jpeg'  # man yuv4mpeg



import math
import random
from typing import List

from alabamaEncode.core.chunk_job import ChunkEncoder


def total_length(scenes: List[ChunkEncoder]) -> int:
    return sum(scene.chunk.length for scene in scenes)


def get_variance(scenes: List[ChunkEncoder]) -> float:
    avg_length = total_length(scenes) / len(scenes)
    variance = sum((scene.chunk.length - avg_length) ** 2 for scene in scenes) / len(
        scenes
    )
    return variance


def annealing(scenes: List[ChunkEncoder], iterations: int) -> List[ChunkEncoder]:

    if len(scenes) < 2:
        return scenes

    current_solution = scenes.copy()
    best_solution = scenes.copy()
    temperature = 1.0
    cooling_rate = 0.95

    for _ in range(iterations):
        neighbor_solution = current_solution.copy()

        # Swap two random scenes
        idx1, idx2 = random.sample(range(len(neighbor_solution)), 2)
        neighbor_solution[idx1], neighbor_solution[idx2] = (
            neighbor_solution[idx2],
            neighbor_solution[idx1],
        )

        # Calculate the variance of current and neighbor solutions
        current_variance = get_variance(current_solution)
        neighbor_variance = get_variance(neighbor_solution)

        # Calculate the change in variance
        delta_variance = neighbor_variance - current_variance

        # Accept the neighbor solution if it's better or according to the temperature and a random probability
        if delta_variance < 0 or random.uniform(0, 1) < math.exp(
            -delta_variance / temperature
        ):
            current_solution = neighbor_solution

        # Update the best solution
        if get_variance(current_solution) < get_variance(best_solution):
            best_solution = current_solution

        # Cool down the temperature
        temperature *= cooling_rate

    return best_solution



import os

from alabamaEncode.parallel_execution.celery_app import app


def worker(concurrency=2):
    print("Starting celery worker")

    if "CELERY_CONCURRENCY" in os.environ:
        concurrency = os.environ["CELERY_CONCURRENCY"]

    app.worker_main(argv=["worker", "--loglevel=info", f"--concurrency={concurrency}"])
    quit()



import copy
import os
import time
from abc import abstractmethod, ABC
from typing import List

from tqdm import tqdm

from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.path import PathAlabama
from alabamaEncode.encoder.codec import Codec
from alabamaEncode.encoder.rate_dist import EncoderRateDistribution
from alabamaEncode.encoder.stats import EncodeStats
from alabamaEncode.metrics.calculate import calculate_metric
from alabamaEncode.metrics.exception import MetricException
from alabamaEncode.metrics.impl.ssim import get_video_ssim
from alabamaEncode.metrics.metric import Metric
from alabamaEncode.metrics.options import MetricOptions


class Encoder(ABC):
    chunk = None
    bitrate = 2000
    crf: int = 23
    passes: int = 1
    video_filters: str = ""
    _output_path: str = None

    @property
    def output_path(self):
        if self._output_path is None:
            return self.chunk.chunk_path

        return self._output_path

    @output_path.setter
    def output_path(self, value):
        self._output_path = value

    speed = 4
    threads = 1
    rate_distribution = EncoderRateDistribution.CQ
    qm_enabled = True
    grain_synth = 3
    qm_min = 8
    qm_max = 15
    tile_cols = -1
    tile_rows = -1
    override_flags: str = ""
    pin_to_core = -1
    niceness = 0

    bit_override = 10

    svt_open_gop = True
    keyint: int = 360  # max chunk is 10s, so this should cover the whole chunk
    svt_sdc: int = 0
    svt_chroma_thing = -2
    svt_supperres_mode = 0
    svt_superres_denom = 8
    svt_superres_kf_denom = 8
    svt_superres_qthresh = 43
    svt_superres_kf_qthresh = 43
    svt_sframe_interval = 0
    svt_sframe_mode = 2
    svt_resize_mode = 0
    svt_resize_denominator = 8
    svt_resize_kf_denominator = 8
    svt_tune = 0  # tune for PsychoVisual Optimization by default
    svt_tf = 1  # temporally filtered ALT-REF frames
    svt_overlay = 0  # enable overlays
    svt_aq_mode = 2  # 0: off, 1: flat, 2: adaptive
    svt_variance_boost_strength = 2
    svt_variance_octile = 6
    svt_enable_variance_boost = 0
    svt_sharpness = 0
    film_grain_denoise: (0 | 1) = 0

    x264_tune = "film"
    x264_mbtree = True
    x264_ipratio = 1.4
    x264_pbratio = 1.3
    x264_aq_strength = -1
    x264_merange = -1
    x264_bframes = -1
    x264_rc_lookahead = -1
    x264_ref = -1
    x264_me = ""
    x264_non_deterministic = False
    x264_subme = -1
    x264_vbv_maxrate = -1
    x264_vbv_bufsize = -1
    x264_slow_firstpass = False
    x264_collect_pass = False

    color_primaries = "bt709"
    transfer_characteristics = "bt709"
    matrix_coefficients = "bt709"
    maximum_content_light_level = ""
    maximum_frame_average_light_level = ""
    chroma_sample_position = 0
    svt_master_display = ""
    hdr = False

    running_on_celery = False

    def supports_float_crfs(self) -> bool:
        return False

    @abstractmethod
    def get_pretty_name(self) -> str:
        pass

    def run(
        self,
        override_if_exists=True,
        timeout_value=-1,
        calcualte_ssim=False,
        metric_to_calculate: Metric = None,
        metric_params: MetricOptions = None,
        on_frame_encoded: callable = None,
    ) -> EncodeStats:
        """
        :param metric_to_calculate: the metric to calculate
        :param calcualte_ssim: self-explanatory
        :param metric_params: dict of vmaf params
        :param override_if_exists: if false and file already exist don't do anything
        :param timeout_value: how much (in seconds) before giving up
        :param on_frame_encoded: callback function that gets called when a frame is encoded,
        with the following parameters: frame: the frame number bitrate: bitrate so far fps: encoding fps
        :return: EncodeStats object with scores bitrate & stuff
        """
        stats = EncodeStats()

        stats.length_frames = self.chunk.get_frame_count()

        should_encode = False

        if not os.path.exists(self.output_path):
            should_encode = True
        elif override_if_exists:
            should_encode = True
        elif not self.chunk.is_done(quiet=True):
            should_encode = True

        if not should_encode:
            tqdm.write("Skipping encode, file already exists")

        if should_encode:
            if self.chunk.path is None or self.chunk.path == "":
                raise Exception("FATAL: output_path is None or empty")

            if not os.path.exists(self.chunk.path):
                raise Exception("FATAL: input file does not exist")
            if self.chunk is None:
                raise Exception("FATAL: chunk is None")
            if self.chunk.chunk_index is None:
                raise Exception("FATAL: current_scene_index is None")

            original_path = copy.deepcopy(self.output_path)
            temp_celery_path = "/tmp/celery/"
            celery_path = (
                f"{temp_celery_path}{os.path.basename(self.output_path)}"
                f"{self.get_chunk_file_extension()}"
            )

            if self.running_on_celery:
                os.makedirs(temp_celery_path, exist_ok=True)
                self.output_path = celery_path

            cli_output = []
            start = time.time()
            commands = self.get_encode_commands()
            self.output_path = original_path

            times_called = 0
            latest_frame_update = 0

            has_frame_callback = (
                self.parse_output_for_output(None) is not None
                and self.passes == 1
                and on_frame_encoded is not None
            )

            try:
                for command in commands:
                    parse_func = None

                    if has_frame_callback:
                        # We can report progress to a callback
                        output_buffer = ""

                        def parse(string):
                            nonlocal output_buffer
                            nonlocal times_called
                            nonlocal latest_frame_update
                            output_buffer += string
                            prog = self.parse_output_for_output(output_buffer)

                            if len(prog) > 0:
                                times_called += 1
                                # tqdm.write("FRAME ENCODED")
                                # tqdm.write(str(prog))
                                latest_frame_update = prog[0]
                                on_frame_encoded(prog[0], prog[1], prog[2])

                                output_buffer = ""

                        parse_func = parse

                    cli_out = (
                        run_cli(
                            command, timeout_value=timeout_value, on_output=parse_func
                        )
                        .verify()
                        .get_output()
                    )
                    cli_output.append(cli_out)

                if self.running_on_celery:
                    # os.rename(celery_path, original_path)
                    # do a copy instead bc "invalid cross-device link"
                    run_cli(f'cp "{celery_path}" "{original_path}"').verify()
                    os.remove(celery_path)

                if has_frame_callback:
                    latest_frame_update = int(latest_frame_update)

                    if latest_frame_update != times_called:
                        for i in range(latest_frame_update - times_called):
                            on_frame_encoded(0, 0, 0)
                            times_called += 1

                stats.time_encoding = time.time() - start

                if (
                    not os.path.exists(self.output_path)
                    or os.path.getsize(self.output_path) < 100
                ):
                    raise Exception(
                        f"FATAL: ENCODE FAILED {self.output_path} NOT FOUND OR TOO SMALL"
                    )

                if stats.time_encoding < 1:
                    stats.time_encoding = 1
            except Exception as e:
                print("Encode command failed, output:")
                for o in cli_output:
                    if isinstance(o, str):
                        o = o.replace("\x08", "")
                        print(o)
                print("Commands: ")
                for c in self.get_encode_commands():
                    print(c)
                raise e

        if metric_to_calculate is not None:
            local_chunk = copy.deepcopy(
                self.chunk
            )  # we need seeking variables from the chunk but the path from the
            # encoder, since the encoder object might have changed the path
            local_chunk.chunk_path = self.output_path
            metric_params = (
                metric_params if metric_params is not None else MetricOptions()
            )
            metric_params.threads = self.threads
            metric_params.video_filters = self.video_filters

            try:
                stats.metric_results = calculate_metric(
                    chunk=local_chunk,
                    options=metric_params,
                    metric=metric_to_calculate,
                )
            except MetricException as e:
                raise Exception(
                    f"{metric_to_calculate} calculation in encoder failed: {e}"
                )

        if calcualte_ssim:
            ssim, ssim_db = get_video_ssim(
                self.output_path,
                self.chunk,
                video_filters=self.video_filters,
                get_db=True,
            )
            stats.ssim = ssim
            stats.ssim_db = ssim_db

        stats.size = os.path.getsize(self.output_path) / 1000
        stats.bitrate = int(
            Ffmpeg.get_total_bitrate(PathAlabama(self.output_path)) / 1000
        )

        return stats

    @abstractmethod
    def get_encode_commands(self) -> List[str]:
        """
        Abstract method overriden by encoders.
        :return: A list of cli commands to encode, according to class fields
        """
        pass

    def get_ffmpeg_pipe_command(self) -> str:
        """
        return cli command that pipes a y4m stream into stdout using the chunk object
        """
        return self.chunk.create_chunk_ffmpeg_pipe_command(
            video_filters=self.video_filters,
            bit_depth=self.bit_override,
        )

    @abstractmethod
    def get_chunk_file_extension(self) -> str:
        return ".mkv"

    @abstractmethod
    def get_version(self) -> str:
        """
        return the version of the encoder
        """
        pass

    def parse_output_for_output(self, buffer) -> [List[str] | None]:
        """
        Parse the output of the encoder and return the frame number, bitrate, and fps.
        :param buffer: The output of the encoder so far
        :return: a list of [frame, bitrate, fps], [] if no output is found, None if not implemented
        """
        return None

    @abstractmethod
    def get_codec(self) -> Codec:
        pass

    def get_crf_range(self) -> [int, int]:
        """
        to be overriden by encoders that support different crf ranges
        """
        return 0, 63

    def clone(self):
        return copy.deepcopy(self)

    def supports_grain_synth(self) -> bool:
        return False



import json


def get_predone():
    obj = json.load(
        open(
            "/home/kokoniara/.config/JetBrains/PyCharm2023.2/scratches/scratch_15.json"
        )
    )
    [c.update({"crf": i}) for i, c in enumerate(obj)]
    # remove ssim total_fps target_miss_proc rate_search_time basename version
    [c.pop("ssim") for c in obj]
    [c.pop("total_fps") for c in obj]
    [c.pop("target_miss_proc") for c in obj]
    [c.pop("rate_search_time") for c in obj]
    [c.pop("basename") for c in obj]
    [c.pop("version") for c in obj]
    [c.pop("chunk_index") for c in obj]
    return obj


if __name__ == "__main__":
    d = get_predone()

    target_vmaf = 95

    def get_score(crf):
        def point_to_score(p):
            """
            calc score including bitrate vmaf and 1% 5% percentiles with weights
            to get the smallest video but with reasonable vmaf and 5% vmaf scores
            """

        for c in d:
            if c["crf"] == crf:
                score = 0

                score_bellow_target_weight = float(0)  # 7
                score_above_target_weight = float(0)  # 4
                score_bitrate_weight = float(0)  # 15
                score_average_weight = float(0)  # 2
                score_5_percentile_target_weight = float(1)  # 5

                # punish if the score is bellow target
                weight = max(0, target_vmaf - c["vmaf"]) * score_bellow_target_weight
                score += weight

                # punish if the score is higher then target
                target_weight = (
                    max(0, c["vmaf"] - target_vmaf) * score_above_target_weight
                )
                score += target_weight

                # how 5%tile frames looked compared to overall score
                # punishing if the video is not consistent
                average_weight = abs(c["vmaf_avg"] - c["vmaf"]) * score_average_weight
                score += average_weight

                # how 5%tile frames looked compared to target, don't if above target
                # punishing if the worst parts of the video are bellow target
                weight_ = (
                    abs(target_vmaf - c["vmaf_percentile_5"])
                    * score_5_percentile_target_weight
                )
                print(crf, weight_)
                score += weight_

                # we punish the hardest for bitrate
                bitrate_weight = max(1, (c["bitrate"] / 100)) * score_bitrate_weight
                score += bitrate_weight  # bitrate

                return score

    def opt_optuna(_get_score, max_probes) -> int:
        import optuna

        def objective(trial):
            value = trial.suggest_int("value", 0, 60)
            return _get_score(value)

        # quiet logs to 0
        # optuna.logging.set_verbosity(0)
        study = optuna.create_study(direction="minimize")
        study.optimize(objective, n_trials=max_probes)
        result = study.best_params["value"]

        return int(result)

    def opt_primitive() -> int:
        closest_score = float("inf")
        best_crf = -1

        for crf in range(15, 60):
            # score = get_score(crf)
            for c in d:
                if c["crf"] == crf:
                    score = abs(target_vmaf - c["vmaf_percentile_5"])
                    print(f"closest_score: {closest_score}, crf: {crf}, score: {score}")
                    if score < closest_score:
                        closest_score = score
                        best_crf = crf
        return best_crf

    # a = opt_optuna(get_score, 2)
    # for c in d:
    #     if c["crf"] == a:
    #         print(c)
    # print(a)
    a = opt_primitive()
    for c in d:
        if c["crf"] == a:
            print(c)
    print(a)

    # plt of vmaf to crf
    import matplotlib.pyplot as plt

    x = [c["crf"] for c in d]
    y = [c["vmaf_percentile_5"] for c in d]

    plt.plot(x, y)
    plt.xlabel("crf")
    plt.ylabel("vmaf_percentile_5")
    # dot where crf touches the line
    plt.plot(a, target_vmaf, "ro")
    plt.show()
    plt.savefig("vmaf_crf.png")



import os
import tempfile
import time
from typing import List

from tqdm import tqdm

from alabamaEncode.core.ffmpeg import Ffmpeg
from alabamaEncode.core.util.bin_utils import get_binary
from alabamaEncode.core.util.cli_executor import run_cli
from alabamaEncode.core.util.path import PathAlabama


class VideoConcatenator:
    def __init__(
        self,
        files: List[str] = None,
        output: str = None,
        file_with_audio: str = None,
        audio_param_override="-c:a libopus -af aformat=channel_layouts=7.1|5.1|stereo -mapping_family 1",
        start_offset=-1,
        end_offset=-1,
        title="",
        encoder_name="TestHoeEncode",
        mux_audio=True,
        subs_file=None,
        audio_only=False,
        temp_dir="",
        copy_included_subs=True,
    ):
        self.files = files
        self.output = output
        self.file_with_audio = file_with_audio
        self.audio_param_override = audio_param_override
        self.start_offset = start_offset
        self.end_offset = end_offset
        self.title = title
        self.encoder_name = encoder_name
        self.mux_audio = mux_audio
        self.subs_file = subs_file
        self.audio_only = audio_only
        self.temp_dir = temp_dir
        self.copy_included_subs = copy_included_subs

        if not self.output:
            print("If muxing please provide an output path")
            raise Exception("No output path not provided")

        if self.temp_dir == "":
            self.temp_dir = os.path.dirname(self.output) + "/"

        self.final_files = []

    def find_files_in_dir(self, folder_path, extension):
        """
        Finds all numbered files in a folder with a given extension
        :param folder_path:  The folder to search
        :param extension: The extension to search for
        :return: self for chaining
        """
        files = []
        for file in os.listdir(folder_path):
            if file.endswith(extension) and "temp.mkv" not in file:
                files.append(os.path.join(folder_path, file))

        files.sort(key=lambda x: int(os.path.splitext(os.path.basename(x))[0]))

        self.files = files
        return self

    def concat_videos(self):
        if not self.files or len(self.files) == 0:
            raise Exception("No files to concat")
        start = time.time()
        self._concat_videos()
        end = time.time()
        print(f"Concat took {end - start} seconds")

    def extract_subs(self, start_offset="", end_offset=""):
        og_file = PathAlabama(self.file_with_audio)
        tracks = Ffmpeg.get_tracks(og_file)
        extracted_subs = []
        sub_counter = 0
        for track in tqdm(
            [track for track in tracks if track["codec_type"] == "subtitle"],
            desc="Extracting Subs",
        ):
            tag_lang = ""
            if "language" in track["tags"]:
                tag_lang = track["tags"]["language"]
            tag_title = ""
            if "title" in track["tags"]:
                tag_title = track["tags"]["title"]

            out_path = f"{self.temp_dir}{tag_lang}.vtt"
            counter = 1
            while os.path.exists(out_path):
                out_path = f"{self.temp_dir}{tag_lang}_{counter}.vtt"
                counter += 1
            try:
                a = (
                    f'{get_binary("ffmpeg")} -v error -y -stats {start_offset} -i {og_file.get_safe()} {end_offset} '
                    f'-map 0:s:{sub_counter} "{out_path}"'
                )
                run_cli(a)
            except Exception as e:
                # print(e)
                pass

            if os.path.exists(out_path):
                extracted_subs.append((out_path, tag_lang, tag_title, sub_counter))
            # else:
            #     print(f"Failed extracing {tag_lang}_{counter} sub")
            #     pass

            sub_counter += 1

        return extracted_subs

    def concat_video_track(self):
        concat_file_path = self.temp_dir + "concat.txt"

        with open(concat_file_path, "w") as f:
            for file in self.files:
                f.write(f"file '{file}'\n")

        concat_vid_ext = ".mkv"

        if Ffmpeg.get_codec(PathAlabama(self.files[0])) == "av1":
            concat_vid_ext = ".ivf"

        self.vid_output = f"{self.temp_dir}vid{concat_vid_ext}"
        print("Concating Video")
        if not os.path.exists(self.vid_output):
            os.system(
                f'{get_binary("ffmpeg")} -y -stats -v error -f concat '
                f'-safe 0 -i "{concat_file_path}" -c:v copy -map_metadata -1 -vsync cfr "{self.vid_output}"'
            )
        if Ffmpeg.check_for_invalid(PathAlabama(self.vid_output)):
            raise Exception("Concating chunks failed")

        os.remove(concat_file_path)

        self.final_files += [f'-i "{self.vid_output}"']

    def encode_audio_tracks(self):
        print("Encoding a audio track")
        self.audio_output = f"{self.temp_dir}audio.mkv"

        vec = [
            get_binary("ffmpeg"),
            "-y",
            "-stats",
            "-v error",
        ]

        if self.start_offset != -1:
            vec += [f"-ss {self.start_offset}"]

        vec += [f'-i "{self.file_with_audio}"']

        if self.end_offset != -1:
            vec += [f"-t {Ffmpeg.get_video_length(PathAlabama(self.vid_output))}"]

        vec += [
            "-map",
            "0:a:0",
            self.audio_param_override,
            "-map_metadata -1",
            f'"{self.audio_output}"',
        ]

        encode_audio = " ".join(vec)
        if not os.path.exists(self.audio_output):
            os.system(encode_audio)
        if Ffmpeg.check_for_invalid(PathAlabama(self.audio_output)):
            raise Exception("Audio Track encoding Failed")

        self.final_files += [f'-i "{self.audio_output}"']

    def has_audio_track(self) -> bool:
        has_audio_track = False
        tracks = Ffmpeg.get_tracks(PathAlabama(self.file_with_audio))
        for track in tracks:
            if track["codec_type"] == "audio":
                has_audio_track = True
                break

        return has_audio_track

    def has_sub_track(self) -> bool:
        has_sub_track = False
        tracks = Ffmpeg.get_tracks(PathAlabama(self.file_with_audio))
        for track in tracks:
            if track["codec_type"] == "subtitle":
                has_sub_track = True
                break

        return has_sub_track

    def _concat_videos(self):
        if os.path.exists(self.output):
            print(f"File {self.output} already exists")
            return

        self.concat_video_track()

        has_audio_track = self.has_audio_track()
        has_sub_track = self.has_sub_track()

        if self.audio_only:
            if not has_audio_track:
                print("No audio track found, not encoding")
                return

            print("Encoding a audio track only")

            print("Encoding a audio track")

            vec = [
                get_binary("ffmpeg"),
                "-y",
                "-stats",
                "-v error",
            ]

            if self.start_offset != -1:
                vec += [f"-ss {self.start_offset}"]

            vec += ["-i", f'"{self.file_with_audio}"']

            if self.end_offset != -1:
                vec += [f"-t {Ffmpeg.get_video_length(PathAlabama(self.vid_output))}"]

            vec += [
                "-map",
                "0:a:0",
                self.audio_param_override,
                "-map_metadata -1",
                f'"{self.output}"',
            ]

            encode_audio = " ".join(vec)
            os.system(encode_audio)
            if Ffmpeg.check_for_invalid(PathAlabama(self.output)):
                print("Invalid file found, exiting")
                return
            os.remove(self.vid_output)
            return

        start_offset_command = ""
        if self.start_offset != -1:
            start_offset_command = f"-ss {self.start_offset}"

        end_offset_command = ""
        if self.end_offset != -1:
            end_offset_command = (
                f"-t {Ffmpeg.get_video_length(PathAlabama(self.vid_output))}"
            )

        if not self.mux_audio or not has_audio_track:
            if not has_audio_track:
                print("No audio track found, not encoding")
            print("Skipping audio")
            os.system(
                f"{get_binary('ffmpeg')} -y -stats -v error -i {self.vid_output} -c copy {self.output}"
            )
            os.remove(self.vid_output)
            if Ffmpeg.check_for_invalid(PathAlabama(self.output)):
                os.remove(self.output)
                raise Exception("VIDEO CONCAT FAILED")
            return

        self.encode_audio_tracks()
        print("Muxing audio into the output")

        title_bit = f' -metadata description="encoded by {self.encoder_name}" '
        if self.title:
            title_bit += f' -metadata title="{self.title}"'

        if self.subs_file is None or self.subs_file[0] == "":
            vec = [
                get_binary("ffmpeg"),
                "-y",
                "-stats",
                "-v error",
                f'-i "{self.vid_output}"',
                f'-i "{self.audio_output}"',
            ]

            sub_tracks = []
            if self.copy_included_subs and has_sub_track:
                sub_tracks = self.extract_subs(
                    start_offset=start_offset_command, end_offset=end_offset_command
                )

                for out_path, tag_lang, tag_title, track_index in sub_tracks:
                    vec += [
                        f'-i "{out_path}"',
                    ]

            if self.encoder_name:
                vec += [f' -metadata description="encoded by {self.encoder_name}"']

            if self.title:
                vec += [f' -metadata title="{self.title}"']

            if self.copy_included_subs and len(sub_tracks) > 0:
                stream_index = 2
                for out_path, tag_lang, tag_title, track_index in sub_tracks:
                    vec += [
                        f"-map {stream_index}:s",
                        f'-metadata:s:s:{track_index} language="{tag_lang}"',
                        f'-metadata:s:s:{track_index} title="{tag_title}"',
                    ]
                    stream_index += 1

            vec += ["-map 0:v", "-map 1:a"]

            if "mp4" in self.output:
                vec += ["-c:s mov_text", "-movflags +faststart"]

            vec += [
                "-map_chapters -1",
                "-c:v copy",
                "-c:a copy",
                "-flags:v +bitexact",
                "-flags:a +bitexact",
                "-avoid_negative_ts make_non_negative",
                "-strict strict",
                "-err_detect compliant",
                "-fflags +genpts+bitexact",
                f'"{self.output}"',
            ]

            final_command = " ".join(vec)
            out = run_cli(final_command).verify().get_output()
            if (
                "Subtitle encoding currently only possible from text to text or bitmap to bitmap"
                in str(out)
            ):
                print("Subtitle encoding failed, trying again")
                for a in vec:
                    if "mov_text" in a or "map 2:s" in a:
                        vec.remove(a)
                final_command = " ".join(vec)
                run_cli(final_command).verify()
        else:
            subs_i = ""
            subs_map = ""
            if len(self.subs_file) > 0 and self.subs_file[0] != "":
                if start_offset_command != "":
                    # offset each track using -itsoffset {offset} -ss {offset}
                    print("Offseting subs")
                    for sub in self.subs_file:
                        temp_sub = f"{sub}.temp.vtt"
                        encode_sub = (
                            f'{get_binary("ffmpeg")} -y -v error -itsoffset {self.start_offset} '
                            f'-ss {self.start_offset} -i "{sub}" "{temp_sub}"'
                        )
                        # print(f"running: {encode_sub}")

                for track_index, sub in enumerate(self.subs_file):
                    if start_offset_command != "":
                        subs_i += f' -i "{sub}.temp.vtt" '
                    else:
                        subs_i += f' -i "{sub}" '
                    subs_map += f"-map {track_index+2} "

            final_command = (
                f'{get_binary("ffmpeg")} -y -stats -v error -i "{self.vid_output}" -i "{self.audio_output}" '
                f'{subs_i} {start_offset_command} -i "{self.file_with_audio}" {end_offset_command} '
                f"{title_bit} -map 0:v -map 1:a {subs_map} -movflags +faststart -map_chapters -1 "
                f'-c:v copy -c:a copy -vsync cfr "{self.output}"'
            )
            os.system(final_command)

        os.remove(self.vid_output)
        os.remove(self.audio_output)

        if Ffmpeg.check_for_invalid(PathAlabama(self.output)):
            os.remove(self.output)
            raise Exception("VIDEO CONCAT FAILED")


def test():
    def create_fake_ivf_and_test(_temp):
        for i in range(20):
            with open(os.path.join(_temp, f"{i}.ivf"), "w") as f:
                f.write(" ")
        # test the file discovery
        vc = VideoConcatenator()
        vc.find_files_in_dir(_temp, ".ivf")
        assert len(vc.files) == 20
        print("Test passed")

    # make temp dir and put 20 empty .ivf files
    temp_dir = tempfile.mkdtemp()
    create_fake_ivf_and_test(temp_dir)

    # make a helpers dir and put 20 empty .ivf files
    sub_dir = os.path.join(temp_dir, "helpers")
    os.mkdir(sub_dir)
    create_fake_ivf_and_test(temp_dir)

    # remove temp dir
    os.system(f"rm -rf {temp_dir}")


if __name__ == "__main__":
    test()
